// Description: a navigator that can grab and move with respect to another node
//
//<b>notes:</b>
//<ul>
//<li> all parent node messages are inherited
//<li> see reset method for default settings
//<li> this node has three modes of operation
//<ul>
//<li> grab - the user is navigated with repect to another node
//<li> move - the user is navigated to another node at maximum speed
//<li> teleport - the user is navigated to another node immediately
//</ul>
//</ul>
//
// Category: User
// Author: Alex Hill
//           11/01/01
// Revision: 07/01/06 Alex Hill - reworked to allow overiding and locked speed to navigator speed 
//
#include <pfcave.h>
#include <ygWorld.h>
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include "grabNavigator.h"

#define DISTANCE_THRESHOLD 0.1
#define HEADING_THRESHOLD 0.5

extern "C" ygNode* construct_grabNavigator(const char* name,bool master) { return new grabNavigator(name,master); }

grabNavigator::grabNavigator(const char* name,bool master) : ygCAVENavigator(name,master)
    {
    setClassName("grabNavigator");
    matrix.makeIdent();
    node = NULL;
    mode = GRAB;
    distanceEasin = 2.0;
    headingEasin = 10.0;
    headingDistance = 20.0;
    override = true;
    addNetMessage("grab",&nodeName,YG_NET_NODE);
    addNetMessage("mode",&mode,YG_NET_INT);
    addNetMessageMap("mode","grab",GRAB);
    addNetMessageMap("mode","move",MOVE);
    addNetMessageMap("mode","teleport",TELEPORT);
    addNetMessage("distanceEasin",&distanceEasin,YG_NET_FLOAT);
    addNetMessage("headingEasin",&headingEasin,YG_NET_FLOAT);
    addNetMessage("headingDistance",&headingDistance,YG_NET_FLOAT);
    addNetMessage("override",&override,YG_NET_BOOL);
    addEvent("grab");
    addEvent("drop");
    addEvent("arrived");
    }

grabNavigator::~grabNavigator(void)
    {
    }

void grabNavigator::reset(void)
    {
    matrix.makeIdent();
    //reset node pointer
    node = NULL;
    //set mode to grab
    mode = GRAB;
    //set the distance easin to 1.0ft
    distanceEasin = 2.0;
    //set the heading easin to 5.0deg
    headingEasin = 10.0;
    //set the heading distance to 5.0ft
    headingDistance = 5.0;
    //set control override to true
    override = true;
    ygCAVENavigator::reset();
    }

void grabNavigator::message(const ygMessage& msg)
    {
    //toggle the grab/drop state with respect to a node
    if (msg == "toggle")
        {
        if (msg.numArgs() == 1)
            {
            ygString currName;
            currName = msg.stringArg(0);
            if (nodeName == currName && node)
                drop();
            else
                {
                nodeName = currName;
                node = ygNodeDB::find(nodeName);
                if (node)
                    grab();
                else
                    msg.error(name(),"(node not found)");
                }
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //grab a node and move with repect to that node
    else if (msg == "grab")
        {
        if (msg.numArgs() == 2)
            {
            nodeName = msg.stringArg(0);
            if (msg.stringArg(1) == "grab")
                mode = GRAB;
            else if (msg.stringArg(1) == "teleport")
                mode = TELEPORT;
            else if (msg.stringArg(1) == "move")
                mode = MOVE;
            else
                mode = mode;
            }
        if (nodeName.length() > 0)
            {
            node = ygNodeDB::find(nodeName);
            if (node)
                grab();
            else
                msg.error(name(),"(node not found)");
            }
        }
    //drop connection with a node or any node grabbed
    else if (msg == "drop")
        {
        if (msg.numArgs() > 0)
            {
            ygString dropName = msg.stringArg(0);
            if (dropName == nodeName)
                drop();
            }
        else
            drop();
        }
    //switch modes or reset to grab mode
    else if (msg == "mode")
        {
        if (msg.numArgs() > 0)
            {
            return;
            }
        }
    //distance at which easin begins
    else if (msg == "distanceEasin")
        {
        return;
        }
    //degrees at which easin begins
    else if (msg == "headingEasin")
        {
        return;
        }
    //distance at which heading changes begin
    else if (msg == "headingDistance")
        {
        return;
        }
    //set override of user control flag
    else if (msg == "override")
        {
        return;
        }
    else
        ygCAVENavigator::message(msg);
    }

void grabNavigator::grab(void)
    {
    //if the node has been found
    if (node)
        {
        /*Now we calculate the relative matrix between the node and this object.
          We use the following formula to calculate our matrix...
                               -1
          matrix = M      * M
                      self     node
        */
        pfMatrix nodeMatrix;
        node->getTransform(nodeMatrix);

        pfMatrix inverseNodeMatrix;
        inverseNodeMatrix.invertOrtho(nodeMatrix);

        pfMatrix selfMatrix;
        getTransform(selfMatrix);

        //calculate the relative matrix
        matrix = selfMatrix*inverseNodeMatrix;
        ygString args("node=");
        args += node->name();
        eventOccurred("grab",args);
        }
    }

void grabNavigator::drop(void)
    {
    //if a node is being tracked then drop
    if (node)
        {
        ygString args("node=");
        args += node->name();
        eventOccurred("drop",args);
        node = NULL;
        nodeName.clear();
        }
    }

void grabNavigator::app(void)
    {
    if (CAVEDistribMaster())
        {
        if ((!checkJoystick() || override)  && node)
            {
            bool arrived = false;
            //if mode is teleport
            if (mode == TELEPORT)
                {
                //get node matrix
                pfMatrix nodeMatrix;
                node->getTransform(nodeMatrix);
                //get node position
                pfVec3 newPos(0,0,0);
                newPos.xformPt(newPos,nodeMatrix);
                //get node heading
                pfVec3 newDir(0,1,0);
                newDir.xformVec(newDir,nodeMatrix);
                float newHeading;
                newHeading = pfArcTan2(-newDir[0],newDir[1]);
                //set new position and heading
                setPosition(newPos);
                setHeading(newHeading);
                arrived = true;
                }
            else if (mode == MOVE)
                {
                //get node matrix
                pfMatrix nodeMatrix;
                node->getTransform(nodeMatrix);
                //get node position
                pfVec3 nodePos(0,0,0);
                nodePos.xformPt(nodePos,nodeMatrix);
                //get node heading
                pfVec3 nodeDir(0,1,0);
                nodeDir.xformVec(nodeDir,nodeMatrix);
                float nodeHeading;
                nodeHeading = pfArcTan2(-nodeDir[0],nodeDir[1]);
                //get delta position
                pfVec3 newPos;
                newPos = position();
                pfVec3 deltaPos;
                deltaPos = nodePos - newPos;
                float distance = deltaPos.length();
                float deltaFactor = 1.0;
                //calculate the new move based on navigation speed
                if (distance > DISTANCE_THRESHOLD)
                    {
                    deltaPos.normalize();
                    if (distance < distanceEasin)
                        deltaFactor = distance/distanceEasin;
                    newPos += deltaPos*deltaFactor*speed()*ygWorld::frameDeltaTime();
                    }
                else 
                    newPos = nodePos;
                float newHeading;
                newHeading = heading();
                float deltaHeading;
                deltaHeading = nodeHeading - newHeading;
                if (deltaHeading > 180.0)
                    deltaHeading -= 360.0;
                if (deltaHeading < -180.0)
                    deltaHeading += 360.0;
                float deltaMagnitude = fabs(deltaHeading);
                //calculate the new heading based on rotation speed
                if (deltaMagnitude > HEADING_THRESHOLD)
                    {
                    if (deltaMagnitude < headingEasin)
                        deltaFactor = deltaMagnitude/headingEasin;
                    else
                        deltaFactor = 1.0;
                    float distanceFactor = 0.0;
                    if (distance < headingDistance)
                        distanceFactor = 1.0 - distance/headingDistance;
                    newHeading += distanceFactor*deltaFactor*(deltaHeading/deltaMagnitude)*rotSpeed()*ygWorld::frameDeltaTime();
                    }
                else 
                    {
                    newHeading = nodeHeading;
                    if (newPos == nodePos)
                        {
                        arrived = true;
                        }
                    }
                //set new position and heading
                setPosition(newPos);
                setHeading(newHeading);
                }
            else 
                {
                //get node matrix
                pfMatrix nodeMatrix;
                node->getTransform(nodeMatrix);
                //get new navigation matrix
                pfMatrix newMatrix;
                newMatrix = matrix*nodeMatrix;
                //get new position
                pfVec3 newPos(0,0,0);
                newPos.xformPt(newPos,newMatrix);
                //get new heading
                pfVec3 newDir(0,1,0);
                newDir.xformVec(newDir,newMatrix);
                float newHeading;
                newHeading = pfArcTan2(-newDir[0],newDir[1]);
                //set new position and heading
                setPosition(newPos);
                setHeading(newHeading);
                }
            if (arrived)
                {
                ygString args("node=");
                args += node->name();
                eventOccurred("arrived",args);
                arrived = false;
                }
            }
        checkGround();
        checkCollision();
        ygNavigator::app();
        }
    else
        ygNode::app();
    }
