// Description: a transform that moves with respect to a wand or other transformed node
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node is useful for grabbing objects and moving them when used in conjunction with a <a href="wandTrigger.html">wandTrigger</a> node
//</ul>
//
// Category: Transformation
// Author: Alex Hill
// Revision: 11/01/01
//
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include "manipulator.h"

using namespace std;

#define YG_POSITION 0
#define YG_ORIENTATION 1
#define YG_SIZE 2
#define YG_SIZE_ALL 3
#define YG_GRAB 4
#define YG_AXIS_AUTO 900
#define YG_AXIS_X 901
#define YG_AXIS_Y 902
#define YG_AXIS_Z 903
#define YG_AXIS_XY 904
#define YG_AXIS_YZ 905
#define YG_AXIS_XZ 906
#define YG_AXIS_XYZ 907

extern "C" ygNode *construct_manipulator(const char *name, bool master) { return new manipulator(name, master);}

manipulator::manipulator(const char *name, bool master) : ygTransform(name, master)
    {
    setClassName("manipulator");
    node = NULL;
    grab_ = NULL;
    mode = YG_POSITION;
    axis = YG_AXIS_AUTO;
    currAxis = YG_AXIS_AUTO;
    transform = NULL;
    lastMatrix.makeIdent();
    addNetMessage("node",&nodeName,YG_NET_NODE);
    addNetMessage("grab",&grabName,YG_NET_NODE);
    addNetMessage("mode",&mode,YG_NET_INT);
    addNetMessageMap("mode","position",YG_POSITION);
    addNetMessageMap("mode","orientation",YG_ORIENTATION);
    addNetMessageMap("mode","size",YG_SIZE);
    addNetMessageMap("mode","sizeall",YG_SIZE_ALL);
    addNetMessageMap("mode","grab",YG_GRAB);
    addNetMessage("axis",&axis,YG_NET_INT);
    addNetMessageMap("axis","auto",YG_AXIS_AUTO);
    addNetMessageMap("axis","x",YG_AXIS_X);
    addNetMessageMap("axis","y",YG_AXIS_Y);
    addNetMessageMap("axis","z",YG_AXIS_Z);
    addNetMessageMap("axis","xy",YG_AXIS_XY);
    addNetMessageMap("axis","yz",YG_AXIS_YZ);
    addNetMessageMap("axis","xz",YG_AXIS_XZ);
    addNetMessageMap("axis","xyz",YG_AXIS_XYZ);
    addEvent("grab");
    addEvent("drop");
    }

manipulator::~manipulator(void)
    {
    }

void manipulator::reset(void)
    {
    //reset last matrix to identity
    lastMatrix.makeIdent();
    //reset node pointer
    node = NULL;
    nodeName.clear();
    //reset grab pointer
    grab_ = NULL;
    grabName.clear();
    //reset mode
    mode = YG_POSITION;
    //reset axis
    axis = YG_AXIS_AUTO;
    currAxis = YG_AXIS_AUTO;
    //create parent transform if necessary
    if (!transform)
        {
        ygString transName(name());
        transName += "xform";
        if (parent() && parent()->name() == transName)
            transform = parent();
        else
            {
            ygNodeDB::createValidName("ygTransform",transName);
            transform = ygNodeDB::create("ygTransform",transName);
            parent()->addChild(transform);
            transform->addChild(this);
            }
        }
    //reset parent transform
    transform->reset();
    ygTransform::reset();
    }


void manipulator::message(const ygMessage& msg)
    {
    //set the node to manipulate
    if (msg == "node")
        {
        if (nodeName.length() > 0)
            {
            node = ygNodeDB::find(nodeName);
            if (node)
                {
                pfMatrix mat;
                node->getParentTransform(mat);
                ((ygTransform*)transform)->setMatrix(mat);
                if (node->isOfClass("ygTransform"))
                    {
                    ((ygTransform*)node)->getMatrix(mat);
                    setMatrix(mat);
                    }
                }
            else
                msg.error(name(),"(node not found)");
            }
        else
            node = NULL;
        }
    //set mode
    else if (msg == "mode")
        {
        return;
        }
    //set axis
    else if (msg == "axis")
        {
        return;
        }
    //toggle the grab/drop state with respect to a node
    else if (msg == "toggle")
        {
        if (msg.numArgs() == 1)
            {
            ygString currName = msg.stringArg(0);
            if (grabName == currName && grab_)
                {
                drop();
                }
            else
                {
                grabName = currName;
                grab_ = ygNodeDB::find(nodeName);
                if (grab_)
                    grab();
                else
                    msg.error(name(),"(node not found)");
                }
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //grab a node and move with repect to that node
    else if (msg == "grab")
        {
        if (grabName.length() > 0)
            {
            grab_ = ygNodeDB::find(grabName);
            if (grab_)
                grab();
            else
                msg.error(name(),"(node not found)");
            }
        }
    //drop connection with a node or any node grabbed
    else if (msg == "drop")
        {
        if (msg.numArgs() == 1)
            {
            ygString dropName = msg.stringArg(0);
            if (dropName == grabName)
                drop();
            }
        else
            drop();
        }
    else
        ygTransform::message(msg);
    }

void manipulator::grab(void)
    {
    //block attempts to manipulate the world
    if (node && node->className() == "ygWorld")
        grab_ = NULL;
    //if the node has been found
    if (grab_)
        {
        //Calculate the relative translation to the grab node
        handle = grab_->origin(this);
        pfMatrix mat;
        getMatrix(mat);
        parentHandle.xformPt(handle,mat);
        //get handle matrix
        pfMatrix handleMat;
        grab_->getTransform(handleMat);
        //get inverse of handle matrix
        pfMatrix handleInverse;
        handleInverse.invertOrtho(handleMat);
        //get current transformation
        pfMatrix currentMat;
        getTransform(currentMat);     
        //calculate difference between handle and current
        handleDiff = currentMat * handleInverse;
        eventOccurred("grab");
        }
    }

void manipulator::drop(void)
    {
    //if a node is being tracked then drop
    if (grab_)
        {
        eventOccurred("drop");
        grab_ = NULL;
        grabName.clear();
        }
    }

void manipulator::app(void)
    {
    if (grab_)
        {
        currAxis = axis;
        if (currAxis == YG_AXIS_AUTO)
            {
            pfVec3 xAxis(1,0,0),yAxis(0,1,0),zAxis(0,0,1);
            float xDot,yDot,zDot;
            xDot = fabs(handle.dot(xAxis));
            yDot = fabs(handle.dot(yAxis));
            zDot = fabs(handle.dot(zAxis));
            if (xDot > yDot && xDot > zDot)
                {
                if (mode == YG_POSITION)
                    currAxis = YG_AXIS_YZ;
                else if (mode == YG_ORIENTATION)
                    currAxis = YG_AXIS_Y;
                else
                    currAxis = YG_AXIS_X;
                }
            else if (yDot > zDot)
                {
                if (mode == YG_POSITION)
                    currAxis = YG_AXIS_XZ;
                else if (mode == YG_ORIENTATION)
                    currAxis = YG_AXIS_Z;
                else
                    currAxis = YG_AXIS_Y;
                }
            else
                {
                if (mode == YG_POSITION)
                    currAxis = YG_AXIS_XY;
                else if (mode == YG_ORIENTATION)
                    currAxis = YG_AXIS_X;
                else
                    currAxis = YG_AXIS_Z;
                }
            }
        pfVec3 currHandle;
        currHandle = grab_->origin(this);
        pfMatrix currMatrix;
        getMatrix(currMatrix);
        pfVec3 currParentHandle;
        currParentHandle.xformPt(currHandle,currMatrix);
        pfMatrix newMatrix;
        getMatrix(newMatrix);
        if (mode == YG_POSITION)
            {
            pfVec3 currVector;
            currVector = currParentHandle - parentHandle;
            if (currAxis == YG_AXIS_X)
                    {
                    currVector[1] = 0;
                    currVector[2] = 0;
                    }
            else if (currAxis == YG_AXIS_Y)
                    {
                    currVector[0] = 0;
                    currVector[2] = 0;
                    }
            else if (currAxis == YG_AXIS_Z)
                    {
                    currVector[0] = 0;
                    currVector[1] = 0;
                    }
            else if (currAxis == YG_AXIS_XY)
                    {
                    currVector[2] = 0;
                    }
            else if (currAxis == YG_AXIS_YZ)
                    {
                    currVector[0] = 0;
                    }
            else if (currAxis == YG_AXIS_XZ)
                    {
                    currVector[1] = 0;
                    }
            pfMatrix delta;
            delta.makeTrans(currVector[0],currVector[1],currVector[2]);
            newMatrix.postMult(delta);
            }
        else if (mode == YG_ORIENTATION)
            {
            pfVec3 size;
            getSize(size);
            pfMatrix sizeMatrix;
            sizeMatrix.makeScale(size[0],size[1],size[2]);
            if (currAxis == YG_AXIS_YZ || currAxis == YG_AXIS_XZ || currAxis == YG_AXIS_XY)
                {
                pfVec3 axis;

                //set the rotation axis
                if (currAxis == YG_AXIS_XZ)
                    axis.set(0,0,1);
                else if (currAxis == YG_AXIS_XY)
                    axis.set(1,0,0);
                else
                    axis.set(0,1,0);
                //get node position vector relative to this node
                pfVec3 rot;
                getOrientation(rot);
                pfMatrix rotMatrix;
                rotMatrix.makeEuler(rot[2],rot[0],rot[1]);
                pfVec3 nodeVector;
                nodeVector.xformPt(currHandle,rotMatrix);
                nodeVector.normalize();
                //get the cross product of the node vector and the rotation axis
                pfVec3 nodeCross;
                nodeCross.cross(axis,nodeVector);
                nodeCross.normalize();
                //establish the object front as the handle vector
                pfVec3 frontAxis(handle);
                pfVec3 frontCross;
                //get the cross product of the front axis and the rotation axis
                frontCross.cross(axis,frontAxis);
                frontCross.normalize();
                float adjacent = frontCross.dot(nodeCross);
                pfVec3 thirdNormal;
                //get a normal towards the node by crossing this cross with the rotation axis
                thirdNormal.cross(axis,frontCross);
                float opposite = thirdNormal.dot(nodeCross);
                //get the rotation to the node vector
                float axisRot = pfArcTan2(opposite,adjacent);
                //create a matrix with this rotation
                newMatrix = sizeMatrix;
                newMatrix.postRot(newMatrix,axisRot,axis[0],axis[1],axis[2]);
                float crossRot;
                //get the rotation about the node cross product
                float delta = axis.dot(nodeVector) - axis.dot(frontAxis);
                if (delta > 1.0)
                    delta = 1.0;
                else if (delta < -1.0)
                    delta = -1.0;
                crossRot = pfArcSin(delta);
                //add this rotation matrix to the existing matrix
                newMatrix.postRot(newMatrix,-crossRot,nodeCross[0],nodeCross[1],nodeCross[2]);
                //add the translation component
                pfVec3 pos;
                getPosition(pos);
                newMatrix.postTrans(newMatrix,pos[0],pos[1],pos[2]);
                }
            else
                {
                pfVec3 tempHandle(handle);
                if (currAxis == YG_AXIS_X)
                    {
                    tempHandle[0] = 0;
                    currHandle[0] = 0;
                    }
                else if (currAxis == YG_AXIS_Y)
                    {
                    tempHandle[1] = 0;
                    currHandle[1] = 0;
                    }
                else if (currAxis == YG_AXIS_Z)
                    {
                    tempHandle[2] = 0;
                    currHandle[2] = 0;
                    }
                currHandle.xformPt(currHandle,sizeMatrix);
                tempHandle.xformPt(tempHandle,sizeMatrix);
                currHandle.normalize();
                tempHandle.normalize();
                newMatrix.preScale(1.0/size[0],1.0/size[1],1.0/size[2],newMatrix);
                pfMatrix delta;
                delta.makeVecRotVec(tempHandle,currHandle);
                newMatrix.preMult(delta);
                newMatrix.preMult(sizeMatrix);
                }
            }
        else if (mode == YG_SIZE)
            {
            pfVec3 currVector;
            currVector.set(currHandle[0]/handle[0],currHandle[1]/handle[1],currHandle[2]/handle[2]);
            if (currAxis == YG_AXIS_X)
                    {
                    currVector[1] = 1;
                    currVector[2] = 1;
                    }
            else if (currAxis == YG_AXIS_Y)
                    {
                    currVector[0] = 1;
                    currVector[2] = 1;
                    }
            else if (currAxis == YG_AXIS_Z)
                    {
                    currVector[0] = 1;
                    currVector[1] = 1;
                    }
            else if (currAxis == YG_AXIS_XY)
                    {
                    currVector[2] = 1;
                    }
            else if (currAxis == YG_AXIS_YZ)
                    {
                    currVector[0] = 1;
                    }
            else if (currAxis == YG_AXIS_XZ)
                    {
                    currVector[1] = 1;
                    }
            pfMatrix delta;
            delta.makeScale(currVector[0],currVector[1],currVector[2]);
            newMatrix.preMult(delta);
            }
        else if (mode == YG_SIZE_ALL)
            {
            pfVec3 currVector;
            currVector.set(currHandle[0]/handle[0],currHandle[1]/handle[1],currHandle[2]/handle[2]);
            if (currAxis == YG_AXIS_X)
                {
                currVector[1] = currVector[0];
                currVector[2] = currVector[0];
                }
            else if (currAxis == YG_AXIS_Y)
                {
                currVector[0] = currVector[1];
                currVector[2] = currVector[1];
                }
            else if (currAxis == YG_AXIS_Z)
                {
                currVector[0] = currVector[2];
                currVector[1] = currVector[2];
                }
            else if (currAxis == YG_AXIS_XY)
                {
                if (currVector[0] > currVector[1])
                    currVector[2] = currVector[0];
                else
                    currVector[2] = currVector[1];
                }
            else if (currAxis == YG_AXIS_YZ)
                {
                if (currVector[1] > currVector[2])
                    currVector[0] = currVector[1];
                else
                    currVector[0] = currVector[2];
                }
            else if (currAxis == YG_AXIS_XZ)
                {
                if (currVector[0] > currVector[2])
                    currVector[1] = currVector[0];
                else
                    currVector[1] = currVector[2];
                }
            else
                {
                if (currVector[0] > currVector[1] && currVector[0] > currVector[2])
                    {
                    currVector[1] = currVector[0];
                    currVector[2] = currVector[0];
                    }
                else if (currVector[1] > currVector[2])
                    {
                    currVector[0] = currVector[1];
                    currVector[2] = currVector[1];
                    }
                else
                    {
                    currVector[0] = currVector[2];
                    currVector[1] = currVector[2];
                    }
                }
            pfMatrix delta;
            delta.makeScale(currVector[0],currVector[1],currVector[2]);
            newMatrix.preMult(delta);
            }
        else if (mode == YG_GRAB)
            {
            //get handle matrix
            pfMatrix handleMat;
            grab_->getTransform(handleMat);
            //get parent transformation
            pfMatrix parentMat;
            getParentTransform(parentMat);
            //invert parent transform
            pfMatrix parentInverse;
            parentInverse.invertOrtho(parentMat);
            // Set the new transformation and return
            newMatrix = handleDiff * handleMat * parentInverse;
            }
        //set the dcs matrix
        setMatrix(newMatrix);
        lastMatrix = newMatrix;
        parentHandle = currParentHandle;
        }
    ygTransform::app();
    }
