// Description: a quaternian based motion path between two positions and orientations
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> the set/value/position message takes a value between 0 and 1
//<li> values outside of the 0 to 1 range will be interpolated accordingly
//</ul>
//
// Category: Transformation
// Author: Alex Hill
// Revision: 11/01/01
//
#include <ygNodeDB.h>
#include <ygWorld.h>
#include <ygNetMessages.h>
#include "mover.h"

extern "C" ygNode* construct_mover(const char* name,bool master) { return new mover(name,master); }

mover::mover(const char* name,bool master) : ygTransform(name,master)
    {
    setClassName("mover");
    position = 0.0;
    posQuat.pos.set(0, 0, 0);
    posQuat.quat.set(0, 0, 0, 1);
    startPosQuat.pos.set(0, 0, 0);
    startPosQuat.ori.set(0, 0, 0);
    startPosQuat.quat.set(0, 0, 0, 1);    
    endPosQuat.pos.set(0, 0, 0);
    endPosQuat.ori.set(0, 0, 0); 
    endPosQuat.quat.set(0, 0, 0, 1);
    dontSaveNetMessage("size");
    addNetMessage("startPosition",&(startPosQuat.pos),YG_NET_VEC3);
    addNetMessage("startOrientation",&(startPosQuat.ori),YG_NET_VEC3);
    addNetMessage("endPosition",&(endPosQuat.pos),YG_NET_VEC3);
    addNetMessage("endOrientation",&(endPosQuat.ori),YG_NET_VEC3);
    addNetMessage("set",&position,YG_NET_FLOAT);
    }

mover::~mover(void)
    {
    }

void mover::reset(void)
    {
    //set current position to zero
    position = 0.0;
    //set current orientation to zero
    posQuat.pos.set(0, 0, 0);
    posQuat.quat.set(0, 0, 0, 1);
    //set start position and orientation to (0,0,0)
    startPosQuat.pos.set(0, 0, 0);
    startPosQuat.ori.set(0, 0, 0);
    startPosQuat.quat.set(0, 0, 0, 1);    
    //set end position and orientation to (0,0,0)
    endPosQuat.pos.set(0, 0, 0);
    endPosQuat.ori.set(0, 0, 0); 
    endPosQuat.quat.set(0, 0, 0, 1);
    ygTransform::reset();
    pfMatrix mat;
    mat.makeQuat(startPosQuat.quat);
    mat.postTrans(mat,startPosQuat.pos[0],startPosQuat.pos[1],startPosQuat.pos[2]);
    setMatrix(mat);
    }

void mover::message(const ygMessage& msg)
    {
    //set the start position of the path
    if (msg == "startPosition")
        {
        return;
        }
    //set the start orientation of the path
    else if (msg == "startOrientation")
        {
        setStartOrientation(startPosQuat.ori);
        }
    //set the end position of the path
    else if (msg == "endPosition")
        {
        return;
        }
    //set the end orientation of the path
    else if (msg == "endOrientation")
        {
        setEndOrientation(endPosQuat.ori);
        }
    //set the position along the motion path
    else if (msg == "set")
        {
        return;
        }
    //set the position along the motion path (deprecated)
    else if (msg == "value" || msg == "position")
        {
        position = msg.floatArg(0);
        }
    else
        ygTransform::message(msg);
    }
    
void mover::setStartPosition(const pfVec3& pos)
    {
    startPosQuat.pos = pos;
    }
        
void mover::setStartOrientation(const pfVec3& rot)
    {
    startPosQuat.ori = rot;
    pfMatrix mat;
    mat.makeEuler(rot[2],rot[0],rot[1]);
    mat.getOrthoQuat(startPosQuat.quat);
    }
    
void mover::setEndPosition(const pfVec3& pos)
    {
    endPosQuat.pos = pos;
    }

void mover::setEndOrientation(const pfVec3& rot)
    {
    endPosQuat.ori = rot;
    pfMatrix mat;
    mat.makeEuler(rot[2],rot[0],rot[1]);
    mat.getOrthoQuat(endPosQuat.quat);
    }

void mover::app(void)
    {
    pfMatrix mat;
    getMatrix(mat);
    pfCoord coord;
    mat.getOrthoCoord(&coord);
    posQuat.pos = coord.xyz;
    mat.getOrthoQuat(posQuat.quat);
    //interpolate the position
    posQuat.pos = startPosQuat.pos + position *
            (endPosQuat.pos - startPosQuat.pos);
    //interpolate the orientation with quaternians
    posQuat.quat.slerp(position,startPosQuat.quat,endPosQuat.quat);
    mat.makeQuat(posQuat.quat);
    mat.postTrans(mat,posQuat.pos[0],posQuat.pos[1],posQuat.pos[2]);
    //set the matrix
    setMatrix(mat);
    ygTransform::app();
    }
