// Description: a node that creates a graph of scene nodes
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> nodes are created in the XY plane
//</ul>
//
// Category: System
// Author: Alex Hill
// Revision: 02/01/06
//
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pr/pfString.h>
#include <Performer/pf/pfDCS.h>
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include <ygWorld.h>
#include "sceneGraph.h"

using namespace std;

extern "C" ygNode* construct_sceneGraph(const char* name,bool master) { return new sceneGraph(name,master); }


sceneGraph::sceneGraph(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("sceneGraph",true);
    node = NULL;
    width = 2.0;
    height = 2.0;
    numNodes = 0;
    numVertices = 10000;
    vertices = (pfVec3*)pfMalloc(numVertices*sizeof(pfVec3), pfGetSharedArena());
    color = (pfVec4*)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
    gstate = new pfGeoState;
    color->set(1,1,1,1);
    addNetMessage("node",&nodeName,YG_NET_NODE);
    addNetMessage("select",&selectName,YG_NET_NODE);
    dontSendNetMessage("select");
    dontSaveNetMessage("select");
    addNetMessage("width",&width,YG_NET_FLOAT);
    addNetMessage("height",&height,YG_NET_FLOAT);
    addEvent("select");
    }

sceneGraph::~sceneGraph(void)
    {
    if (nodes.size() > 0)
        pfnode()->removeChild(nodes[0]);
    for (int i=0; i<nodes.size(); i++)
        {
        while (nodes[i]->getNumChildren() > 0)
            nodes[i]->removeChild(nodes[i]->getChild(0));
        }
    for (int i=0; i<nodes.size(); i++)
        pfDelete(nodes[i]);
    for (int i=0; i<connections.size(); i++)
        pfDelete(connections[i]);
    pfDelete(vertices);
    pfDelete(color);
    pfDelete(gstate);
    }


void sceneGraph::reset(void)
    {
    resetNetMessage("width");
    resetNetMessage("height");
    node = NULL;
    numNodes = 0;
    ygNode::reset();
    }


void sceneGraph::message(const ygMessage& msg)
    {
    //set root node name
    if (msg == "node")
        {
        if (nodeName.length() > 0)
            {
            if (nodeName != lastNode)
                {
                node = ygNodeDB::find(nodeName);
                if (node)
                    {
                    currNum = 0;
                    currVertex = 0;
                    currConn = 0;
                    buildGraph(node);
                    numNodes = currNum;
                    }
                else
                    msg.error(name(),"(node not found)");
                }
            }
        else
            {
            for (int i=0; i<nodes.size(); i++)
                {
                while (nodes[i]->getNumChildren() > 0)
                    nodes[i]->removeChild(nodes[i]->getChild(0));
                }
            numNodes = 0;
            }
        lastNode = nodeName;
        }
    //set the node to select
    else if (msg == "select")
        {
        checkSelect();
        }
    //set the color of the text
    else if (msg == "width")
        {
        return;
        }
    //set the color of the text
    else if (msg == "height")
        {
        return;
        }
    else
        ygNode::message(msg);
    }

float sceneGraph::buildGraph(ygNode* currNode)
    {
    int parentNum = currNum;
    if (parentNum < nodes.size())
        {
        while (nodes[parentNum]->getNumChildren() > 0)
             nodes[parentNum]->removeChild(nodes[parentNum]->getChild(0));
        nodes[parentNum]->addChild(currNode->debugText());
        }
    else
        {
        pfDCS* dcs = new pfDCS;
        dcs->addChild(currNode->debugText());
        nodes.push_back(dcs);
        if (parentNum == 0)
            pfnode()->addChild(dcs);
        }
    currNum++;
    int numLines;
    float parentWidth;
    parentWidth = 2*(currNode->debugString()->getBBox()->max[0]+YG_DEBUG_PADDING);
    if (currNode->numChildren() == 0 || currVertex + 2*(currNode->numChildren()+2) >= numVertices)
        return parentWidth;
    else if (currNode->numChildren() > 1)
        numLines = currNode->numChildren() + 2;
    else
        numLines = 2;
    if (currConn >= connections.size())
        connections.push_back(createConnection(numLines));
    int parentConn = currConn;
    connections[parentConn]->getGSet(0)->setNumPrims(numLines);
    connections[parentConn]->getGSet(0)->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, (void*)(&vertices[currVertex]), NULL);
    int parentVertex = currVertex;
    currVertex += 2*numLines;
    currConn++;
    float offset = 0.0;
    float connWidth = 0.0;
    float childWidth = 0.0;
    float firstOffset;
    //call buildGraph recursively on all children
    for (int i=0; i<currNode->numChildren(); i++)
        {
        if (i > 0)
            {
            offset += childWidth/2;
            offset += width;
            connWidth += width;
            }
        int childNum = currNum;
        childWidth = buildGraph(currNode->child(i));
        offset += childWidth/2;
        if (i == 0)
            firstOffset = offset;
        connWidth += childWidth;
        vertices[parentVertex].set(offset,-height/2,0.0);
        parentVertex++;
        vertices[parentVertex].set(offset,-height+0.2+YG_DEBUG_PADDING,0.0);
        parentVertex++;
        nodes[parentNum]->addChild(nodes[childNum]);
        }
    vertices[parentVertex].set(connWidth/2,-0.2-YG_DEBUG_PADDING,0.0);
    parentVertex++;
    vertices[parentVertex].set(connWidth/2,-height/2,0.0);
    parentVertex++;
    if (numLines > 2)
        {
        vertices[parentVertex].set(firstOffset,-height/2,0.0);
        parentVertex++;
        vertices[parentVertex].set(offset,-height/2,0.0);
        parentVertex++;
        }
    for (int i=0; i<2*numLines; i++)
        vertices[parentVertex-1-i][0] -= connWidth/2;
    for (int i=0; i<currNode->numChildren(); i++)
        ((pfDCS*)(nodes[parentNum]->getChild(i+1)))->setTrans(vertices[parentVertex-2*numLines+2*i][0],-height,0.0);
    nodes[parentNum]->addChild(connections[parentConn]);
    pfSphere sphere;
    sphere.center.set(0,0,0);
    sphere.radius = connWidth/2;
    connections[parentConn]->setBound(&sphere,PFBOUND_STATIC);
    if (parentWidth > connWidth)
         return parentWidth;
    return connWidth;
    }

pfGeode* sceneGraph::createConnection(int numLines)
    {
    pfGeoSet* gset = new pfGeoSet;
    gset->setPrimType(PFGS_LINES);
    gset->setNumPrims(numLines);
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, (void*)vertices, NULL);
    gstate->setMode(PFSTATE_ENWIREFRAME,1);
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,color,NULL);
    pfBox box;
    box.min.set(-1000,-1000,-1000);
    box.max.set(1000,1000,1000);
    gset->setBound(&box,PFBOUND_STATIC);
    pfGeode* geode = new pfGeode;
    geode->addGSet(gset);
    return geode;
    }

void sceneGraph::checkSelect(void)
    {
    if (selectName.length() > 0 && selectName != lastSelect)
        {
        pfVec3 pos(0,0,0);
        for (int i=0; i<numNodes; i++)
            {
            if (strcmp(((pfGroup*)(nodes[i]->getChild(0)))->getChild(1)->getName(),selectName.c_str()) == 0)
                {
                pfDCS* dcs = nodes[i];
                pfMatrix mat;
                mat.makeIdent();
                pfMatrix dcsmat;
                while(dcs != pfnode())
                    {
                    dcs->getMat(dcsmat);
                    mat *= dcsmat;
                    dcs = (pfDCS*)dcs->getParent(0);
                    }
                pos.xformPt(pos,mat);
                }
            }
        ygString args;
        char value[32];
        args += " xpos=";
        sprintf(value,"%f",pos[0]);
        args += value;
        args += " ypos=";
        sprintf(value,"%f",pos[1]);
        args += value;
        //report the position of the selected graph node
        eventOccurred("select",args);
        }
    lastSelect = selectName;
    }

void sceneGraph::app(void)
    {
    if (node && node->isChanged(YG_CHANGE_TREE))
        {
        //rebuild graph
        currNum = 0;
        currVertex = 0;
        currConn = 0;
        buildGraph(node);
        numNodes = currNum;
        }
    ygNode::app();
    }
