// Description: creates a rising plume of smoke
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> you must add the following to your .ygrc file in order to use this node:
//<ul>
//<li> setenv YG_PF_INIT_SMOKES
//<li> setenv YG_PRELOAD_CLASSES smoke
//</ul>
//</ul>
//
// Category: Geometry
// Author: Josheph Tremonti
// Revision: 11/01/01
//
#include <Performer/pf/pfTraverser.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfGeode.h>
#include <math.h>
#include <ygNetMessages.h>
#include <ygWorld.h>
#include "smoke.h"

extern "C" ygNode* construct_smoke(const char* name,bool master) { return new smoke(name,master); }

typedef struct _smokeData
    {
    pfVec3 eye;
    } smokeData_t;

smoke::smoke(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("smoke",true);
    createSmoke();
    density.set(1.0,1.0,0.5);
    velocity.set(3.0,0.2);
    color.xyz.set(1.0,1.0,1.0);
    color.hpr.set(1.0,1.0,1.0);
    type_ = PFUSMOKE_SMOKE;
    active = true;
    pfuSmokeVelocity(smoke_,velocity[1],velocity[0]);
    pfuSmokeDensity(smoke_,density[0],density[1],density[2]);
    pfVec3 origin(0,0,0);
    pfuSmokeOrigin(smoke_,origin,1.0f);
    pfVec3 direction(0,0,1);
    pfuSmokeDir(smoke_,direction);
    pfuSmokeDuration(smoke_,0xffffffff);
    pfuSmokeColor(smoke_,color.xyz,color.hpr);
    pfuSmokeType(smoke_,type_);
    pfuSmokeMode(smoke_,PFUSMOKE_START);
    addNetMessage("velocity",&velocity,YG_NET_VEC2);
    addNetMessage("density",&density,YG_NET_VEC3);
    addNetMessage("color",&color,YG_NET_COORD);
    addNetMessage("off",&active,YG_NET_BOOL_NEGATE);
    addNetMessage("mode",&type_,YG_NET_INT);
    addNetMessageMap("mode","missile",PFUSMOKE_MISSILE);
    addNetMessageMap("mode","explosion",PFUSMOKE_EXPLOSION);
    addNetMessageMap("mode","fire",PFUSMOKE_FIRE);
    addNetMessageMap("mode","smoke",PFUSMOKE_SMOKE);
    addNetMessageMap("mode","dust",PFUSMOKE_DUST);
    }

smoke::~smoke(void)
    {
    }

void smoke::reset(void)
    {
    resetNetMessage("velocity");
    resetNetMessage("density");
    resetNetMessage("color");
    resetNetMessage("off");
    resetNetMessage("mode");
    pfuSmokeVelocity(smoke_,velocity[1],velocity[0]);
    pfuSmokeDensity(smoke_,density[0],density[1],density[2]);
    pfVec3 origin(0,0,0);
    pfuSmokeOrigin(smoke_,origin,1.0f);
    pfVec3 direction(0,0,1);
    pfuSmokeDir(smoke_,direction);
    pfuSmokeDuration(smoke_,0xffffffff);
    pfuSmokeColor(smoke_,color.xyz,color.hpr);
    pfuSmokeType(smoke_,type_);
    pfuSmokeMode(smoke_,PFUSMOKE_START);
    ygNode::reset();
    }

void smoke::message(const ygMessage& msg)
    {
    //set smoke mode
    if (msg == "mode")
        {
        pfuSmokeType(smoke_,type_);
        }
    //set the smoke density, dissipation, and expansion
    else if (msg == "density")
        {
        pfuSmokeDensity(smoke_,density[0],density[1],density[2]);
        }
    //set the smoke velocity and turbulence
    else if (msg == "velocity")
        {
        pfuSmokeVelocity(smoke_,velocity[1],velocity[0]);
        }
    //set the smoke begining and ending color
    else if (msg == "color")
        {
        pfuSmokeColor(smoke_,color.xyz,color.hpr);
        }
    //turn the smoke on
    else if (msg == "on")
        {
        if (!active)
            {
            active = true;
            pfuSmokeMode(smoke_,PFUSMOKE_START);
            netMessageChanged("off");
            }
        }
    //turn the smoke off
    else if (msg == "off")
        {
        if (active)
            pfuSmokeMode(smoke_,PFUSMOKE_START);
        else
            pfuSmokeMode(smoke_,PFUSMOKE_STOP);
        }
    else
        ygNode::message(msg);
    }

void smoke::createSmoke()
    {
    pfSphere bsph;
    bsph.radius = 100.0f;
    bsph.center.set(0,0,0);
    pfGeode* geode = new pfGeode;
    geode->setBound(&bsph,PFBOUND_STATIC);
    geode->setTravFuncs(PFTRAV_DRAW,smokePreDraw,smokePostDraw);
    data_ = (smokeData_t*) pfMalloc(sizeof(smokeData_t),pfGetSharedArena());
    data_->eye.set(0,0,0);
    geode->setTravData(PFTRAV_DRAW,data_);
    geode->setTravMask(PFTRAV_ISECT,0,PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
    smoke_ = pfuNewSmoke();
    pfnode()->addChild(geode);
    }

void smoke::app(void)
    {
    ygNode* viewer = ygWorld::world()->viewer();
    if (viewer)
        {
        data_->eye = viewer->origin(this);
        data_->eye.normalize(); 
        }
    }
    
int smoke::smokePreDraw(pfTraverser*,void* data)
    {
    //cast the traversal data as the smokeData structure
    smokeData_t* smokeData = (smokeData_t*)data;
    glDepthMask(GL_FALSE);
    pfuDrawSmokes(smokeData->eye);
    return PFTRAV_CONT;
    }

int smoke::smokePostDraw(pfTraverser*,void* data)
    {
    glDepthMask(GL_TRUE);
    return PFTRAV_CONT;
    }
