// Description: creates a node that will record the microphone input to a file
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> the volume of a sound is the region in which the sound will play at 100% amplitude(loudness)
//<li> falloffDistance is a region outside the volume in which the amplitude is diminished as the user moves further away from the source(volume)
//<li> recorded sounds are stored in .aiff (Audio Interchange File Format)
//<li> the sound server, snerd, must be run on the a local or remote machine
//<li> if the sound file does not exist in YG_PATH, then it will be created in the first directory in the path
//<li> sounds under a switch will turn off but must be restarted when the switch is toggled back on
//</ul>
//
// Category: Sound
// Author: Alex Hill
// Revision: 08/01/03
//
#include <pfcave.h>
#include <bergenRecord.h>
#include <ygWorld.h>
#include "ygNetMessages.h"
#include "soundRecorder.h"

using namespace std;

extern "C" ygNode* construct_soundRecorder(const char* name,bool master) { return new soundRecorder(name,master); }

struct _soundRecorderPrivateData
    {
    ygString filename;
    bergenRecord* sample;
    bool isRecording;
    bool netRecording;
    bool isPaused;
    float curAmplitude;
    float distance;
    pfVec3 currPosition;
    float updateInterval, lastUpdate;
    };


soundRecorder::soundRecorder(const char* name,bool master) : ygSoundServer(name,master)
    {
    setClassName("soundRecorder",true);
    p_ = new struct _soundRecorderPrivateData;
    p_->sample = NULL;
    p_->netRecording = false;
    p_->isRecording = false;
    p_->curAmplitude = 0;
    p_->updateInterval = 0.1;
    if (getenv("YG_SOUND_INTERVAL"))
        p_->updateInterval = atof(getenv("YG_SOUND_INTERVAL"));
    p_->lastUpdate = 0;
    addNetMessage("record", &p_->netRecording, YG_NET_BOOL_NEGATE);
    addNetMessage("stop", &p_->netRecording, YG_NET_BOOL);
    dontSaveNetMessage("stop");
    addNetMessage("file", &p_->filename, YG_NET_FILE);
    addNetMessage("updateInterval",&p_->updateInterval,YG_NET_FLOAT);
    }

soundRecorder::~soundRecorder(void)
    {
    if (p_->sample)
        delete p_->sample;
    delete p_;
    }


void soundRecorder::reset(void)
    {
    //reset recording flag to false
    resetNetMessage("record");
    p_->isRecording = false;
    //set update interval to 0.1
    resetNetMessage("updateInterval");
    //if a sample exists then delete it
    if (p_->sample)
        {
        stop();
        delete p_->sample;
        p_->sample = NULL;
        }
    ygSoundServer::reset();
    }


void soundRecorder::message(const ygMessage& msg)
    {
    //set the name of the sound file
    if (msg == "file")
        {
        setFile(p_->filename);
        }
    //start the sound recording
    else if (msg == "record")
        {
        if (p_->netRecording)
            record();
        else
            stop();
        }
    //stop the recording sound
    else if (msg == "stop")
        {
        if (p_->netRecording)
            record();
        else
            stop();
        }
    //set the update interval for sound server
    else if (msg == "updateInterval")
        {
        return;
        }
    else
        ygSoundServer::message(msg);
    }


void soundRecorder::setFile(const ygString& val)
    {
    if (p_->sample)
        {
        stop();
        delete p_->sample;
        p_->sample = NULL;
        }
    p_->filename = val;
    }



void soundRecorder::setDistance(float val)
    {
    if (p_->sample)
        p_->sample->setDistance(val);
    p_->distance = val;
    }


void soundRecorder::setAmplitude(float val)
    {
    if (p_->sample)
        p_->sample->setAmplitude(val);
    ygSoundServer::setAmplitude(val);
    }


void soundRecorder::setFalloffDistance(float val)
    {
    if (p_->sample)
        p_->sample->setFalloffDistance(val);
    ygSoundServer::setFalloffDistance(val);
    }


void soundRecorder::setAttenuation(int val)
    {
    if (p_->sample)
        p_->sample->setAttenuationType((AttenuationType)val);
    ygSoundServer::setAttenuation(val);
    }


bool soundRecorder::isPlaying(void)
    {
    return p_->isRecording;
    }


bool soundRecorder::isPaused(void)
    {
    return p_->isPaused;
    }


void soundRecorder::record(void)
    {
    if (p_->isRecording)
        return;
    //if no sample then create a new bergenRecord
    if (!p_->sample)
        p_->sample = new bergenRecord(p_->filename.c_str(),server());
    p_->sample->setAmplitude(amplitude());
    p_->sample->setFalloffDistance(falloffDistance());
    p_->sample->setAttenuationType((AttenuationType)attenuation());
    setDistance(computeDistance());
    //record the sample
    p_->sample->record();
    //generate a heartbeat for the server
    heartbeat();
    p_->isRecording = true;
    eventOccurred("start");
    }


void soundRecorder::pause(void)
    {
    if (p_->isRecording)
        {
        p_->isPaused = true;
        p_->isRecording = false;
        p_->sample->pause();
        }
    }


void soundRecorder::stop(bool sendStopCommand)
    {
    //if a sample is playing
    if (p_->isRecording || p_->isPaused)
        {
        //send a stop command to the server
        if ((sendStopCommand) && (p_->sample))
            p_->sample->stop();
        p_->isRecording = false;
        p_->netRecording = false;
        eventOccurred("stop");
        }
    }


void soundRecorder::app(void)
    {
    //if a sample exists
    if (p_->sample)
        {
        //generate a heartbeat for the server
        heartbeat();
        //if the sample is playing
        if (p_->isRecording)
            {
            //if update interval has passed then update sound server
            if ((ygWorld::frameTime() - p_->lastUpdate) > p_->updateInterval)
                {
                float newDistance = computeDistance();
                if (newDistance != p_->distance)
                    setDistance(newDistance);
                //determine position of this sound
                pfVec3 newPos;
                newPos = origin();
                pfVec3 cavePos;
                //get position relative to CAVE
                CAVENavConvertWorldToCAVE(newPos.vec,cavePos.vec);
                if (cavePos != p_->currPosition)
                    {
                    if (p_->sample)
                        p_->sample->setPosition(cavePos.vec[0],cavePos.vec[2],-cavePos.vec[1]);
                    p_->currPosition = cavePos;
                    }
                p_->lastUpdate = ygWorld::frameTime();
                }
            }
        }
    ygSoundServer::app();
    }
