// Description: a tri-strip surface of variable size
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node is affected by scene lighting
//</ul>
//
// Category: Geometry
// Author: Alex Hill
//           11/01/02
//           02/01/06 Alex Hill - changed logic to create a 2x2 surface by default
// Revision: 07/01/06 Alex Hill - added functionality to create a parent ygTransform when necessary
//
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfGeode.h>
#include <ygPFTextureCache.h>
#include <ygWorld.h>
#include <ygNetMessages.h>
#include "ygNodeDB.h"
#include "surface.h"

using namespace std;

extern "C" ygNode* construct_surface(const char* name,bool master) { return new surface(name,master); }

surface::surface(const char* name,bool master) : ygGeometry(name,master)
    {
    setClassName("surface",true);
    xSize_ = 2;
    ySize_ = 2;
    lastXSize_ = 0;
    lastYSize_ = 0;
    colors = NULL;
    texture = NULL;
    gset = NULL;
    gstate = NULL;
    transform = NULL;
    //create geode
    geode = new pfGeode;
    geode->setName(name);
    object->addChild(geode);
    addNetMessage("vertices",&xSize_,YG_NET_INT);
    addNetMessage("vertices",&ySize_,YG_NET_INT);
    dontSendNetMessage("vertices");
    addNetMessage("file",&textureFile,YG_NET_FILE);
    addNetMessage("texture",&textureFile,YG_NET_FILE);
    dontSaveNetMessage("texture");
    }

surface::~surface(void)
    {
    }

void surface::reset(void)
    {
    //reset number of vertices to 2
    resetNetMessage("vertices");
    createGeometry(xSize_,ySize_);
    //reset texture file name
    resetNetMessage("file");
    if (!transform && parent())
        {
        ygString transName(name());
        transName += "xform";
        if (parent()->name() == transName)
            transform = parent();
        }
    ygGeometry::reset();
    }


void surface::message(const ygMessage& msg)
    {
    //determine the number of vertices
    if (msg == "vertices") 
        {
        if (msg.numArgs() == 1)
            ySize_ = xSize_;
        createGeometry(xSize_,ySize_);
        }
    //set the surface texture
    else if (msg == "file")
        {
        setTexture(textureFile);
        }
    //set the surface texture (deprecated)
    else if (msg == "texture") 
        {
        setTexture(textureFile);
        }
    //create a parent ygTransform and pass the message
    else if (msg == "position" || msg == "orientation" || msg == "size" || msg == "updateInterval" || msg == "reliableKey")
        {
        createTransform();
        transform->parseMessage(msg);
        }
    else
        ygGeometry::message(msg);
    }

void surface::createGeometry(int xSize, int ySize)
    {
    if (xSize < 2)
        xSize = 2;
    if (ySize < 2)
        ySize = 2;
    //if geode has been created
    if (xSize != lastXSize_ || ySize != lastYSize_)
        {
        if (gset)
            {
            //delete gset
            gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, NULL, NULL);
            gset->setAttr(PFGS_NORMAL3, PFGS_PER_VERTEX, NULL, NULL);
            gset->setAttr(PFGS_COLOR4, PFGS_PER_VERTEX, NULL, NULL);
            gset->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, NULL, NULL);
            pfDelete(gset->getPrimLengths());
            gset->setPrimLengths(NULL);  
            pfDelete(verts);
            pfDelete(norms);
            pfDelete(colors);
            pfDelete(texCoords);
            pfDelete(ilist);
            }
        else
            {
            //create geoset
            gset = new pfGeoSet;
            gset->setPrimType(PFGS_TRISTRIPS);
            //create geostate
            gstate = new pfGeoState;
            mat = new pfMaterial();;
            gstate->setAttr(PFSTATE_FRONTMTL, mat);
            gstate->setAttr(PFSTATE_BACKMTL, mat);
            gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
            gset->setGState(gstate);
            geode->addGSet(gset);
            }

        //allocate vertex, normal, color, and texture coordinate arrays
        verts = (pfVec3*) pfMalloc(xSize * ySize * sizeof(pfVec3), pfGetSharedArena());
        norms = (pfVec3*) pfMalloc(xSize * ySize * sizeof(pfVec3), pfGetSharedArena());
        colors = (pfVec4*) pfMalloc(xSize * ySize * sizeof(pfVec4), pfGetSharedArena());
        texCoords = (pfVec2*) pfMalloc(xSize * ySize * sizeof(pfVec2), pfGetSharedArena());

        //set up vertex, normal, color, and texture corrdinate values
        int index;
        for (int y = 0, index = 0; y < ySize; y++)
            {
            for (int x = 0; x < xSize; x++, index++) 
                {
                verts[index].set(x/(float)(xSize-1) - 0.5,y/(float)(ySize-1) - 0.5,0.0);
                norms[index].set(0, 0, 1);
                colors[index].set(1,1,1,1);
                texCoords[index].set(x/(float)(xSize-1), y/(float)(ySize-1));
                }
            }

        int* lengths = (int*)pfMalloc((ySize-1)*sizeof(int), pfGetSharedArena());
        for (int y = 0; y < ySize-1; y++)
            lengths[y] = xSize * 2;
        gset->setPrimLengths(lengths);
        gset->setNumPrims(ySize-1);

        //set up indexing
        ilist = (ushort*)pfMalloc(2*xSize*(ySize-1)*sizeof(ushort), pfGetSharedArena());
        index = 0;
        for (int y=0; y<ySize-1; y++)
            {
            for (int x=0; x<xSize; x++)
                {
                ilist[index++] = x + y*xSize + xSize;
                ilist[index++] = x + y*xSize;
                }
            }

        //add array attributes to geoset
        gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, ilist);
        gset->setAttr(PFGS_NORMAL3, PFGS_PER_VERTEX, norms, ilist);
        gset->setAttr(PFGS_COLOR4, PFGS_PER_VERTEX, colors, ilist);
        gset->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, texCoords, ilist);               

        xSize_ = xSize;
        ySize_ = ySize;
        lastXSize_ = xSize_;
        lastYSize_ = ySize_;
        netMessageChanged("vertices");
        setChanged(YG_CHANGE_GEOMETRY);
        }
    }

void surface::setTexture(const ygString& file)
    {
    textureFile = file;
    if (textureFile.length() > 0)
        texture = ygPFTextureCache::textureReference(file.c_str());
    else
        texture = NULL;
    if (texture && gstate)
        {
        int comp;
        //get number of image components
        texture->getImage(NULL,&comp,NULL,NULL,NULL);
        //if alpha channel then set transparency
        if (comp == 4)
            {
            gstate->setMode(PFSTATE_TRANSPARENCY,PFTR_BLEND_ALPHA);
            gset->setDrawBin(PFSORT_TRANSP_BIN);
            }
        else
            {
            gstate->setMode(PFSTATE_TRANSPARENCY,PFTR_OFF);
            gset->setDrawBin(PFSORT_OPAQUE_BIN);
            }
        gstate->setAttr(PFSTATE_TEXTURE, texture);
        gstate->setMode(PFSTATE_ENTEXTURE, PF_ON);
        setChanged(YG_CHANGE_TEXTURE);
        }
    }


void surface::createTransform(void)
    {
    if (!transform)
        {
        ygString transName(name());
        transName += "xform";
        ygNodeDB::createValidName("ygTransform",transName);
        transform = ygNodeDB::create("ygTransform",transName);
        parent()->addChild(transform);
        transform->addChild(this);
        transform->reset();
        }
    }
