// Description: a variable size array of values
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node generates an event when any one of the array values changes
//<li> any other event generated by this node will include value arguments
//<li> the reference message is used to reference other value nodes
//</ul>
//
// Category: Math
// Author: Alex Hill
//           11/01/01
// Revision: 01/15/03 Alex Hill - added event message
//
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include "value.h"

using namespace std;

extern "C" ygNode* construct_value(const char* name,bool master) { return new value(name,master); }

value::value(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("value");
    setValueSize(1);
    values[0] = 0.0;
    lastValues[0] = 0.0;
    size = 1;
    setSize(deltas,1,0);
    deltas[0] = 1.0;
    integer = false;
    reference = NULL;
    addNetMessage("reference",&refNode,YG_NET_NODE);
    addNetMessage("values",&values,YG_NET_VECTOR);
    addNetMessage("deltas",&deltas,YG_NET_VECTOR);
    addNetMessage("minimums",&minimums,YG_NET_VECTOR);
    addNetMessage("maximums",&maximums,YG_NET_VECTOR);
    addNetMessage("integer",&integer,YG_NET_BOOL_NEGATE);
    addEvent("changed");
    }
    
value::~value(void)
    {
    }

void value::reset(void)
    {
    //set size to 1
    setValueSize(1);
    //initialize value to 0.0
    values[0] = 0.0;
    lastValues[0] = 0.0;
    //set delta size to 1
    setSize(deltas,1,deltas.size());
    //initialize delta value to 1.0
    deltas[0] = 1.0;
    //set minimums size to 0
    setSize(minimums,0,minimums.size());
    //set maximums size to 0
    setSize(maximums,0,maximums.size());
    reference = NULL;
    refNode.clear();
    //set integer flag to false
    integer = false;
    ygNode::reset();
    }

void value::message(const ygMessage& msg)
    {
    //make this node an instance of another node
    if (msg == "reference")
        {
        if (msg.numArgs() > 0 && className() == "value")
            {
            refNode = msg.stringArg(0);
            ygNode* node = ygNodeDB::find(refNode);
            if (node)
                {
                if (node->isOfClass("value"))
                    reference = (value*)node;
                else
                    reference = NULL;
                }
            else
                msg.error(name(),"(node not found)");
            }
        else
            reference = NULL;
        }
    //else, if referenced, send the message to the reference node
    else if (reference)
        {
        reference->parseMessage(msg);
        }
    //set the value of a single array element
    else if (msg == "set" || msg == "value")
        {
        if (msg.numArgs() > 0)
            {
            float newValue = msg.floatArg(0);
            if (msg.numArgs() > 1) 
                {
                int element = msg.intArg(1);
                if (element >= 0)
                    {    
                    if (element >= size)
                        setValueSize(element+1);
                    values[element] = newValue;
                    }
                else
                    msg.error(name(),"(index out of range)");
                }
            else
                {
                if (size > 0)
                    values[0] = newValue;
                }
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //set the values of all array elements
    else if (msg == "values")
        {
        if (values.size() != size)
            setValueSize(values.size());
        }
    //set the size of the array
    else if (msg == "size")
        {
        if (msg.numArgs() == 1)
            {
            int newSize = msg.intArg(0);
            if (newSize >= 0)
                setValueSize(newSize);
            else
                msg.error(name(),"(index out of range)");
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //set the value of a single delta element
    else if (msg == "delta")
        {
        if (msg.numArgs() > 0)
            {
            float newValue = msg.floatArg(0);
            if (msg.numArgs() > 1) 
                {
                int element = msg.intArg(1);
                if (element >= 0)
                    {    
                    if (element >= deltas.size())
                        setSize(deltas,element+1,deltas.size());
                    deltas[element] = newValue;
                    }
                else
                    msg.error(name(),"(index out of range)");
                }
            else
                deltas[0] = newValue;
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //set the values of all delta array elements
    else if (msg == "deltas")
        {
        return;
        }
    //set the value of single minimum clamp
    else if (msg == "minimum")
        {
        if (msg.numArgs() > 0)
            {
            float newValue = msg.floatArg(0);
            if (msg.numArgs() > 1) 
                {
                int element = msg.intArg(1);
                if (element >= 0)
                    {    
                    if (element >= minimums.size())
                        setSize(minimums,element+1,minimums.size());
                    minimums[element] = newValue;
                    }
                else
                    msg.error(name(),"(index out of range)");
                }
            else
                {
                if (minimums.size() == 0)
                    setSize(minimums,1,minimums.size());
                minimums[0] = newValue;
                }
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //set the values of all minimum array elements
    else if (msg == "minimums")
        {
        return;
        }
    //set the value of single maximum clamp
    else if (msg == "maximum")
        {
        if (msg.numArgs() > 0)
            {
            float newValue = msg.floatArg(0);
            if (msg.numArgs() > 1) 
                {
                int element = msg.intArg(1);
                if (element >= 0)
                    {    
                    if (element >= maximums.size())
                        setSize(maximums,element+1,maximums.size());
                    maximums[element] = newValue;
                    }
                else
                    msg.error(name(),"(index out of range)");
                }
            else
                {
                if (maximums.size() == 0)
                    setSize(maximums,1,maximums.size());
                maximums[0] = newValue;
                }
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //set the values of all maximum array elements
    else if (msg == "maximums")
        {
        return;
        }
    //decrement all or a single element by delta
    else if (msg == "decrement")
        {
        if (msg.numArgs() > 0)
            {
            int element = msg.intArg(0);
            if (element >= 0 && element < size)
                {
                if (element >= deltas.size())
                    values[element] -= deltas[deltas.size()-1];
                else
                    values[element] -= deltas[element];
                }
            else
                msg.error(name(),"(index out of range)");
            }
        else
            {
            for (int i=0;i<size;i++)
                {
                if (i >= deltas.size())
                    values[i] -= deltas[deltas.size()-1];
                else
                    values[i] -= deltas[i];
                }
            }
        }
    //increment all or a single element by delta
    else if (msg == "increment")
        {
        if (msg.numArgs() > 0)
            {
            int element = msg.intArg(0);
            if (element >= 0 && element < size)
                {
                if (element >= deltas.size())
                    values[element] += deltas[deltas.size()-1];
                else
                    values[element] += deltas[element];
                }
            else
                msg.error(name(),"(index out of range)");
            }
        else
            {
            for (int i=0;i<size;i++)
                {
                if (i >= deltas.size())
                    values[i] += deltas[deltas.size()-1];
                else
                    values[i] += deltas[i];
                }
            }
        }
    //set if the value should be rounded to the nearest integer
    else if (msg == "integer")
        {
        return;
        }
    //generate an event with associated arguments
    else if (msg == "event")
        {
        if (msg.numArgs() > 0)
            {
            ygString eventName = msg.stringArg(0);
            ygString args;
            char index[32];
            //for each value in the array
            for (int i=0;i<size;i++)
                {
                if (i == 0)
                    {
                    args += "value=";
                    sprintf(index,"%g",values[i]);
                    args += index;
                    }
                args += " ";
                args += "value";
                sprintf(index,"%d",i);
                args += index;
                args += "=";
                sprintf(index,"%g",values[i]);
                args += index;
                }
            if (msg.numArgs() == 1)
                {
                //add $value and $valueX arguments to each event
                eventOccurred(eventName,args);
                }
            else
                {
                for (int i=1;i<msg.numArgs();i++)
                    {
                    args += " ";
                    args += msg.stringArg(i);
                    }
                //add all $value arguments to each event plus any arguments
                eventOccurred(eventName,args);
                }
            }
        }
    else
        ygNode::message(msg);
    }

void value::setValueSize(int newSize)
    {
    setSize(values,newSize,values.size());
    setSize(lastValues,newSize,lastValues.size());
    size = newSize;
    }
    
void value::setSize(vector<float>& array, int newSize, int oldSize, float initValue)
    {
    //if size has changed
    if (newSize != oldSize)
        {
        array.resize(newSize);
        if (newSize > 0)
            {
            //if old size is 0 then fill with initial values
            if (oldSize == 0)
                {
                for (int i=0;i<newSize;i++) 
                    array[i] = initValue;
                }
            //else, if new size is larger then pad with initial values
            else if (newSize > oldSize)
                {
                for (int i=oldSize;i<newSize;i++) 
                    array[i] = initValue;
                }
            }
        }
    }

bool value::isChanged(void) 
    { 
    bool changed = false; 
    for (int i=0;i<size;i++)
        {
        if (values[i] != lastValues[i])
            changed = true;
        }
    return changed;
    }
    
void value::clampValues(void)
    {
    //for each value in the array
    for (int i=0;i<size;i++)
        {
        //if exceeding a limit then clamp
        if (minimums.size() > 0)
            {
            int minIndex = i;
            if (minIndex >= minimums.size())
                minIndex = minimums.size()-1;
            if (values[i] < minimums[minIndex])
                values[i] = minimums[minIndex];
            }
        if (maximums.size() > 0)
            {
            int maxIndex = i;
            if (maxIndex >= maximums.size())
                maxIndex = maximums.size()-1;
            if (values[i] > maximums[maxIndex])
                values[i] = maximums[maxIndex];
            }
        if (integer)
            values[i] = rint(values[i]);
        }
    }

void value::app(void)
    {
    //clamp values
    clampValues();
    //if changed then generate event
    if (isChanged())
        {
        ygString args;
        char index[32];
        //for each value in the array
        for (int i=0;i<size;i++)
            {
            if (i == 0)
                {
                args += "value=";
                if (integer)
                    sprintf(index,"%d",(int)values[i]);
                else
                    sprintf(index,"%f",values[i]);
                args += index;
                }
            args += " value";
            sprintf(index,"%d",i);
            args += index;
            args += "=";
            if (integer)
                sprintf(index,"%d",(int)values[i]);
            else
                sprintf(index,"%f",values[i]);
            args += index;
            }
        //for each value in the array
        for (int i=0;i<size;i++)
            lastValues[i] = values[i];
        //generate changed event with $value, $valueX where X=index
        eventOccurred("changed",args);
        }
    ygNode::app();
    }


void value::appTraverse(void)
    {
    //start timing
    timingBegin();
    //call appTraverse recursively on all children
    for (int i=0; i < numChildren(); i++)
        child(i)->appTraverse();
    //execute app method
    app();
    //end timing
    timingEnd();
    }

float value::getValue(int element) const
    {
    if (!reference)
        {
        if (element < size)
            return values[element];
        else
            return 0.0;
        }
    else
        return reference->getValue(element);
    }

float value::getLastValue(int element) const
    {
    if (!reference)
        {
        if (element < size)
            return lastValues[element];
        else
            return 0.0;
        }
    else
        return reference->getLastValue(element);
    }

void value::setValue(const float& iValue, int element)
    {
    if (!reference)
        { 
        if (element >= size)
            setValueSize(element+1);
        values[element] = iValue;
        }
    else
        reference->setValue(iValue,element);
    }

int value::getSize(void) const
    {
    if (!reference)
        return size;
    else
        return reference->getSize();
    }

void value::createDebugNode(void)
    {
    pfGeode* geode = new pfGeode;
    if (!geode)
        {
        cerr << "ERROR: ygTransform::createDebugGeode: failed to create pfGeode\n";
        return;
        }
    pfGeoSet *gset = new pfGeoSet;
    int *lengths = (int *) pfMalloc(6 * sizeof(int),pfGetSharedArena());
    gset->setPrimType(PFGS_POLYS);
    gset->setNumPrims(6);
    lengths[0] = 4;
    lengths[1] = 4;
    lengths[2] = 4;
    lengths[3] = 4;
    lengths[4] = 4;
    lengths[5] = 4;
    gset->setPrimLengths(lengths);
    pfVec3 * verts = (pfVec3 *) pfMalloc(8 * sizeof(pfVec3),pfGetSharedArena());
    pfBox box;
    box.min.set(-0.5,-0.125,-0.25);
    box.max.set(0.5,0.125,0.25);
    verts[0] = box.min;
    verts[1][0] = box.max[0]; verts[1][1] = box.min[1]; verts[1][2] = box.min[2];
    verts[2][0] = box.max[0]; verts[2][1] = box.max[1]; verts[2][2] = box.min[2];
    verts[3][0] = box.min[0]; verts[3][1] = box.max[1]; verts[3][2] = box.min[2];
    verts[4][0] = box.min[0]; verts[4][1] = box.min[1]; verts[4][2] = box.max[2];
    verts[5][0] = box.max[0]; verts[5][1] = box.min[1]; verts[5][2] = box.max[2];
    verts[6] = box.max;
    verts[7][0] = box.min[0]; verts[7][1] = box.max[1]; verts[7][2] = box.max[2];
    ushort *ilist_v = (ushort *) pfMalloc(24 * sizeof(ushort),pfGetSharedArena());
    ilist_v[0] = 0; ilist_v[1] = 3; ilist_v[2] = 2; ilist_v[3] = 1;
    ilist_v[4] = 0; ilist_v[5] = 1; ilist_v[6] = 5; ilist_v[7] = 4; 
    ilist_v[8] = 1; ilist_v[9] = 2; ilist_v[10] = 6; ilist_v[11] = 5;
    ilist_v[12] = 2; ilist_v[13] = 3; ilist_v[14] = 7; ilist_v[15] = 6;
    ilist_v[16] = 3; ilist_v[17] = 0; ilist_v[18] = 4; ilist_v[19] = 7;
    ilist_v[20] = 4; ilist_v[21] = 7; ilist_v[22] = 6; ilist_v[23] = 5;
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, ilist_v);
    ushort *ilist_c = (ushort *) pfMalloc(1 * sizeof(ushort),pfGetSharedArena());
    ilist_c[0] = 0;
    gset->setAttr(PFGS_COLOR4, PFGS_OVERALL, debugColor(), ilist_c);
    gset->setGState(new pfGeoState);
    gset->setLineWidth(1.0);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    gset->getGState()->setMode(PFSTATE_ENWIREFRAME,1);
    geode->addGSet(gset); 
    gset = new pfGeoSet;
    gset->setPrimType(PFGS_LINES);
    gset->setNumPrims(10);
    verts = (pfVec3 *) pfMalloc(14 * sizeof(pfVec3),pfGetSharedArena());
    verts[0][0] = -0.333333; verts[0][1] = 0; verts[0][2] = 0.167777;
    verts[1][0] = -0.333333; verts[1][1] = 0; verts[1][2] = -0.167777;
    verts[2][0] = -0.167777; verts[2][1] = 0; verts[2][2] = 0.167777;
    verts[3][0] = 0; verts[3][1] = 0; verts[3][2] = 0.167777;
    verts[4][0] = -0.167777; verts[4][1] = 0; verts[4][2] = 0;
    verts[5][0] = 0; verts[5][1] = 0; verts[5][2] = 0;
    verts[6][0] = -0.167777; verts[6][1] = 0; verts[6][2] = -0.167777;
    verts[7][0] = 0; verts[7][1] = 0; verts[7][2] = -0.167777;
    verts[8][0] = 0.167777; verts[8][1] = 0; verts[8][2] = 0.167777;
    verts[9][0] = 0.333333; verts[9][1] = 0; verts[9][2] = 0.167777;
    verts[10][0] = 0.2; verts[10][1] = 0; verts[10][2] = 0;
    verts[11][0] = 0.333333; verts[11][1] = 0; verts[11][2] = 0;
    verts[12][0] = 0.167777; verts[12][1] = 0; verts[12][2] = -0.167777;
    verts[13][0] = 0.333333; verts[13][1] = 0; verts[13][2] = -0.167777;
    ilist_v = (ushort *) pfMalloc(20 * sizeof(ushort),pfGetSharedArena());
    ilist_v[0] = 0; ilist_v[1] = 1; ilist_v[2] = 2; ilist_v[3] = 3;
    ilist_v[4] = 3; ilist_v[5] = 5; ilist_v[6] = 4; ilist_v[7] = 5; 
    ilist_v[8] = 4; ilist_v[9] = 6; ilist_v[10] = 6; ilist_v[11] = 7;
    ilist_v[12] = 8; ilist_v[13] = 9; ilist_v[14] = 10; ilist_v[15] = 11;
    ilist_v[16] = 12; ilist_v[17] = 13; ilist_v[18] = 9; ilist_v[19] = 13;
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, ilist_v);
    gset->setAttr(PFGS_COLOR4, PFGS_OVERALL, debugColor(), ilist_c);
    gset->setGState(new pfGeoState);
    gset->setLineWidth(1.0);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    gset->getGState()->setMode(PFSTATE_ENWIREFRAME,1);
    geode->addGSet(gset); 
    setDebugNode((pfGroup*)geode);
    }
