// Description: the base class from which all scripted Ygdrasil nodes are derived
//
//<b>notes:</b>
//<ul>
//<li> see reset method for default settings
//<li> this node can also be created with the aliases: node, ygGroup, and group
//<li> handles basic node functions such as maintaining child/parent pointers, handling events, and processing initial messages
//<li> this node can be used to group items in a scene
//</ul>
//
// Category: Foundation
// Author: Dave Pape
//           11/01/01
//           10/01/02 Alex Hill - added pfChanged and setPFChanged
//           10/10/02 Alex Hill - changed event message to take multiple arguments
//           09/01/04 Alex Hill - added "create" message
//           06/28/05 Alex Hill - added functions related to maintaining debug geometry
//           02/01/06 Alex Hill - added addEvent, setTransform, and other functions related to debug display
// Revision: 07/01/06 Alex Hill - revised pfChanged logic to avoid concurrency problems
//
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <Performer/pfdu.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfString.h>
#include <Performer/pr/pfMaterial.h>
#include <Performer/pf/pfText.h>
#include <Performer/pf/pfSCS.h>
#include <iostream>
#include <vector>
#include <queue>
#include <map>
#include "ygNodeDB.h"
#include "ygNet.h"
#include "ygNetMessages.h"
#include "ygWorld.h"
#include "ygDebugFlags.h"
#include "ygEvent.h"
#include "ygEventMessages.h"
#include "ygUtil.h"
#include "ygNode.h"

using namespace std;


struct _ygNodePrivateData
    {
    ygString name;
    ygString className;
    ygString classDerivation;
    ygString netClassName;
    vector<ygMessage*> initMessages;
    ygEventMessages eventMessages;
    ygDebugFlags debug;
    vector<ygNode*> children;
    vector<ygString*> comments;
    ygNode* parent;
    pfGroup* pfnode;
    bool isMaster;
    uint changes;
    uint changeAdd;
    uint changeRemove;
    bool debugNetMessages;
    bool debugMessages;
    bool debugDisplay;
    bool debugSelected;
    pfGroup* debugNode;
    pfGroup* debugText;
    pfString* debugString;
    pfVec4* debugColor;
    ygNetMessages netMessages;
    ygString netChildren;
    ygString netMessage;
    ygString resetEvent;
    int resetNumber;
    int frameCount, eventCount;
    double frameStart;
    double frameTimeTotal;
    double appTimeTotal;
    };



extern "C" ygNode* construct_ygNode(const char* name,bool master) { return new ygNode(name,master); }


ygNode::ygNode(const char* name,bool master,bool createPfNode)
    {
    p_ = new struct _ygNodePrivateData;
    p_->isMaster = master;
    setClassName("ygNode",true);
    p_->name = name;
    p_->eventMessages.setNode(this);
    p_->parent = NULL;
    //if createPfNode flag set then create a new group node
    if (createPfNode)
        setPfNode(new pfGroup);
    else
        setPfNode(NULL);
    //print out all messages sent over the network
    debugFlag("netMessages",&p_->debugNetMessages);
    //print out all messages received or sent by an event
    debugFlag("messages",&p_->debugMessages);
    //display debug outline
    debugFlag("display", &p_->debugDisplay);
    //change debug outline color to selected
    debugFlag("selected", &p_->debugSelected);
    p_->debugNode = NULL;
    p_->debugColor = (pfVec4*) pfMalloc(sizeof(pfVec4),pfGetSharedArena());
    p_->debugColor->set(YG_DEBUG_COLOR);
    p_->debugString = NULL;
    p_->debugText = NULL;
    p_->netMessages.setNodeName(p_->name);
    //distribute list of children nodes
    addNetMessage("addChild",&p_->netChildren,YG_NET_STRING_ADD);
    dontSendNetMessage("addChild");
    dontSaveNetMessage("addChild");
    addNetMessage("removeChild",&p_->netChildren,YG_NET_STRING_DELETE);
    dontSendNetMessage("removeChild");
    dontSaveNetMessage("removeChild");
    //distribute network messages
    addNetMessage("message",&p_->netMessage,YG_NET_STRING);
    dontSendNetMessage("message");
    dontSaveNetMessage("message");
    //distribute display debug status
    addNetMessage("edit",&p_->debugDisplay,YG_NET_BOOL_NEGATE);
    addNetMessage("selected",&p_->debugSelected,YG_NET_BOOL_NEGATE);
    dontSaveNetMessage("selected");
    //register updated event
    addEvent("updated");
    p_->frameCount = 0;
    p_->eventCount = 0;
    p_->frameTimeTotal = 0;
    p_->appTimeTotal = 0;
    p_->changes = YG_CHANGE_NONE;
    p_->changeAdd = YG_CHANGE_NONE;
    p_->changeRemove = YG_CHANGE_NONE;
    }


ygNode::~ygNode(void)
    {
    vector<ygMessage*>::const_iterator iter;
    for (iter=p_->initMessages.begin(); iter != p_->initMessages.end();
         ++iter)
        delete *iter;
    ygNodeDB::remove(this);
    if (p_->pfnode)
        pfDelete(p_->pfnode);
    delete p_;
    }


void ygNode::reset(void)
    {
    //reset display debug
    resetNetMessage("edit");
    //reset display debug selected
    resetNetMessage("selected");
    //clear event messages
    clearEventMessages();
    //send intitial messages
    sendInitMessages();
    }


void ygNode::sendInitMessages(void)
    {
    vector<ygMessage*>::const_iterator iter;
    for (iter=p_->initMessages.begin(); iter != p_->initMessages.end(); ++iter)
        ygWorld::world()->scheduleMessage(**iter);
    }


void ygNode::message(const ygMessage& msg)
    {
    //configure messages to send in response to an event
    if (msg == "when")
        {
        int i;
        if (msg.numArgs() < 2)
            msg.error(name(),"(wrong number of arguments)");
        ygString eventName;
        eventName = msg.stringArg(0);
        for (i=1; i < msg.numArgs(); i++)
            when(eventName, msg.stringArg(i));
        }
    //add the given node as a child
    else if (msg == "addChild" || msg == "removeChild")
        {
        removeAllChildren();
        ygString* childName;
        int i=0;
        while (childName = p_->netChildren.nextToken(" ",&i))
            {
            ygNode* n = ygNodeDB::find(*childName);
            if (n)
                addChild(n,true);
            else
                {
                cerr << "WARNING: " << msg << " - could not find child node" << endl;
                int j;
                p_->netChildren.clear();
                for (j=0; j < numChildren(); j++)
                    {
                    if (j>0)
                        p_->netChildren += " ";
                    p_->netChildren += child(j)->name();
                    }
                }
            delete childName;
            }
        netMessageChanged("addChild");
        }
    //turn debuging on and off
    else if (msg == "debug")
        {
        if (msg.numArgs() > 0)
            {
            bool val;
            ygString flag = msg.stringArg(0);
            if (msg.numArgs() > 1)
                val = msg.boolArg(1);
            else
                val = true;
            setDebug(flag,val);
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //turn display debuging on and off
    else if (msg == "edit")
        {
        updateDebug();
        }
    //turn display selected debuging on and off
    else if (msg == "selected")
        {
        updateDebug();
        }
    //print the arguments to standard output
    else if (msg == "print")
        {
        int i;
        ygString data;
        for (i=0; i < msg.numArgs(); i++)
            {
            data += msg.stringArg(i);
            data += " ";
            }
        cout << name() << ": " << data << endl;
        }
    //reset and send initial and/or individual event messages
    else if (msg == "reset")
        {
        if (msg.numArgs() > 0)
            {
            bool val = msg.boolArg(0);
            if (val)
                {
                if (msg.numArgs() > 1)
                    {
                    ygString eventName = msg.stringArg(1);
                    if (msg.numArgs() > 1)
                        {
                        int eventNum = msg.intArg(2);
                        clearEventMessages(eventName,eventNum);
                        }
                    else
                        clearEventMessages(eventName);
                    }
                else
                    reset();
                }
            }
        else
            reset();
        }
    //generate an event with arguments of the form "arg=val"
    else if (msg == "event")
        {
        if (msg.numArgs() > 0)
            {
            ygString eventName = msg.stringArg(0);
            if (msg.numArgs() > 1)
                {
                ygString args;
                for (int i=1;i<msg.numArgs();i++)
                    {
                    args += msg.stringArg(i);
                    args += " ";
                    }
                //process event messages with arguments
                eventOccurred(eventName,args);
                }
           else
                {
                //process event messages without arguments
                eventOccurred(eventName);
                }
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //make a network key unreliable
    else if (msg == "unreliablekey")
        {
        if (msg.numArgs() > 0)
            unreliableMessage(msg.stringArg(0));
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //make a network key reliable
    else if (msg == "reliablekey")
        {
        if (msg.numArgs() > 0)
            reliableMessage(msg.stringArg(0));
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //reset the tree from this node down
    else if (msg == "resetTree")
        resetTree();
    //create a new child node with optional name
    else if (msg == "create")
        {
        ygNode* node=NULL;
        if (msg.numArgs() > 0)
            {
            ygString name;
            ygString type;
            type = msg.stringArg(0);
            if (msg.numArgs() > 1)
                {
                name = msg.stringArg(1);
                ygNodeDB::createValidName(type,name);
                node = ygNodeDB::create(type,name);
                }
            else
                {
                ygNodeDB::createValidName(type,name);
                node = ygNodeDB::create(type,name);
                }
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        if (node)
            {
            addChild(node);
            node->reset();
            cout << name() << ": created node " << node->name() << endl;;
            }
        }
    //receive net messages
    else if (msg == "message")
        {
        if (isNetMaster())
            {
            if (p_->debugNetMessages)
                cout << name() << " got net message '" << p_->netMessage << "'\n";
            ygMessage newMsg;
            newMsg.parseString(p_->netMessage,this);
            if (newMsg.recipient() == this)
                parseMessage(newMsg);
            else if (newMsg.recipient())
                ygWorld::world()->scheduleMessage(newMsg);
            else
                ygWorld::world()->scheduleMessage(newMsg,true);
            }
        }
    else
        cout << "WARNING: " << name() << " received unknown message " << msg << endl;
    }


void ygNode::resetTree(void)
    {
    //reset this node
    reset();
    //call resetTree recursively on all children
    for (int i=0; i < numChildren(); i++)
        child(i)->resetTree();
    }


void ygNode::appTraverse(void)
    {
    //start timing
    timingBegin();
    //execute app method
    app();
    //call appTraverse recursively on all children
    for (int i=0; i < numChildren(); i++)
        child(i)->appTraverse();
    //end timing
    timingEnd();
    }


void ygNode::app(void)
    {
    ygNet::midframeProcess();
    p_->changes &= ~p_->changeRemove;
    p_->changes |= p_->changeAdd;
    p_->changeRemove = YG_CHANGE_NONE;
    p_->changeAdd = YG_CHANGE_NONE;
    timingAppDone();
    }


ygNode* ygNode::parent(void) const
    {
    return p_->parent;
    }


int ygNode::numChildren(void) const
    {
    return p_->children.size();
    }


ygNode* ygNode::child(int index) const
    {
    if ((index >= 0) && (index < p_->children.size()))
        return p_->children[index];
    else
        return NULL;
    }


void ygNode::addChild(ygNode* child,bool doLocal)
    {
    if (!child)
        {
        cerr << "ERROR (ygNode::addChild()): attempted to add a NULL"
            " pointer as a child of " << name() << endl;
        }
    //if net master or doLocal is set
    else if (isNetMaster() || doLocal)
        {
        pfMatrix mat;
        bool update=false;
        //if child has a parent have it remove the child
        if (child->parent())
            {
            if (child->pfnode()->isOfType(pfSCS::getClassType()))
                {
                child->getTransform(mat);
                update = true;
                }
            child->parent()->removeChild(child,doLocal);
            }
        //set the parent of the child to this node
        child->p_->parent = this;
        //bubble up the changes mask
        p_->changeAdd |= child->p_->changes;
        //push the child to the back of children vector
        p_->children.push_back(child);
        //establish parent/child relationship for associated pfnodes
        if ((pfnode()) && (child->pfnode()))
            pfnode()->addChild(child->pfnode());
        if (isNetMaster() && !doLocal)
            updateChildrenMessage();
        if (update)
            child->setTransform(mat);
        //set and bubble up child changed mask
        child->setChanged(YG_CHANGE_TREE);
        }
    //else, generate an addChild message and send
    else
        {
        ygString netmsg("addChild(");
        netmsg += child->name();
        netmsg += ")";
        sendNetMessage(netmsg);
        }
    }


void ygNode::removeChild(ygNode* child,bool doLocal)
    {
    //if net master or doLocal is set
    if (isNetMaster() || doLocal)
        {
        vector<ygNode*>::iterator i = p_->children.begin();
        //while searching children
        while ((i != p_->children.end()) && (*i != child))
            ++i;
        //if child is found
        if (i != p_->children.end())
            {
            //remove child
            p_->children.erase(i);
            child->p_->parent = NULL;
            //undo parent/child relationship for associated pfnodes
            if ((pfnode()) && (child->pfnode()))
                pfnode()->removeChild(child->pfnode());
            if (isNetMaster())
                updateChildrenMessage();
            //set changed mask
            setChanged(YG_CHANGE_TREE);
            }
        }
    //else, generate a removeChild message and send
    else
        {
        ygString netmsg("removeChild(");
        netmsg += child->name();
        netmsg += ")";
        sendNetMessage(netmsg);
        }
    }


void ygNode::setClassName(const ygString& s,bool net)
    {
    p_->className = s;
    p_->classDerivation += ":";
    p_->classDerivation += s;
    p_->classDerivation += ":";
    p_->debug.checkFlags(s);
    if (net)
        p_->netClassName = s;
    }


const ygString& ygNode::className(void) const
    {
    return p_->className;
    }


const ygString& ygNode::classDerivation(void) const
    {
    return p_->classDerivation;
    }


bool ygNode::isOfClass(const ygString& className) const
    {
    ygString colonName(":");
    colonName += className;
    colonName += ":";
    return p_->classDerivation.contains(colonName);
    }


const ygString& ygNode::netClassName(void) const
    {
    return p_->netClassName;
    }


const ygString& ygNode::name(void) const
    {
    return p_->name;
    }


void ygNode::addInitMessage(const ygString& str)
    {
    ygMessage* msg = new ygMessage;
    msg->parseString(str,this);
    p_->initMessages.push_back(msg);
    }


void ygNode::clearInitMessages(void)
    {
    p_->initMessages.clear();
    }


const vector<ygMessage*>& ygNode::getInitMessages(void)
    {
    return p_->initMessages;
    }


const vector<ygMessage*>& ygNode::getNetMessages(bool all)
    {
    return p_->netMessages.getMessages(all);
    }


const vector<ygString>& ygNode::getNetMessageMap(const ygString& msg)
    {
    return p_->netMessages.getMessageMap(msg);
    }


void ygNode::setPfNode(pfGroup* n)
    {
    p_->pfnode = n;
    if (n)
        n->setName(name().c_str());
    }


void ygNode::setChanged(uint change)
    {
    p_->changeAdd |= change;
    p_->changeAdd |= YG_CHANGE_STATUS;
    if (p_->parent)
        p_->parent->setChanged(change);
    }


bool ygNode::isChanged(uint change)
    {
    if (change & p_->changes)
        {
        p_->changeRemove |= change;
        p_->changeRemove |= YG_CHANGE_STATUS;
        return true;
        }
    else
        return false;
    }


pfGroup* ygNode::pfnode(void) const
    {
    return p_->pfnode;
    }


pfVec3 ygNode::origin(const ygNode* node) const
    {
    pfVec3 origin(0,0,0), pos, retval;
    pfMatrix mat;
    getTransform(mat);
    pos.xformPt(origin,mat);
    //if node is given get position relative to the node
    if (node)
        {
        pfMatrix mat,invmat;
        node->getTransform(mat);
        invmat.invertAff(mat);
        retval.xformPt(pos,invmat);
        return retval;
        }
    //else, get absolute position of this node
    else
        return pos;
    }


void ygNode::getTransform(pfMatrix& mat) const
    {
    getParentTransform(mat);
    if (pfnode()->isOfType(pfSCS::getClassType()))
        {
        pfMatrix scsmat;
        ((pfSCS*)pfnode())->getMat(scsmat);
        scsmat *= mat;
        mat = scsmat;
        }
    }


void ygNode::setTransform(const pfMatrix& mat)
    {
    return;
    }


void ygNode::getParentTransform(pfMatrix& mat) const
    {
    pfMatrix scsmat;
    ygNode* node = parent();
    mat.makeIdent();
    while (node)
        {
        if (node->pfnode()->isOfType(pfSCS::getClassType()))
            {
            ((pfSCS*)(node->pfnode()))->getMat(scsmat);
            mat *= scsmat;
            }
        node = node->parent();
        }
    }


/* removeAllChildren() is intended solely for use when we receive
   a new list of children from the network */
void ygNode::removeAllChildren(void)
    {
    vector<ygNode*>::iterator i = p_->children.begin();
    for ( ; (i != p_->children.end()); ++i)
        {
        (*i)->p_->parent = NULL;
        if ((pfnode()) && ((*i)->pfnode()))
            pfnode()->removeChild((*i)->pfnode());
        }
    p_->children.clear();
    }


void ygNode::updateChildrenMessage(void)
    {
    int i;
    p_->netChildren.clear();
    for (i=0; i < numChildren(); i++)
        {
        if (i>0)
            p_->netChildren += " ";
        p_->netChildren += child(i)->name();
        }
    netMessageChanged("addChild");
    }

/***** Event functions *****/


void ygNode::addEvent(const ygString& event)
    {
    p_->eventMessages.addEvent(event);
    }


void ygNode::when(const ygString& event,const ygString& message)
    {
    p_->eventMessages.addMessage(event,message);
    }


void ygNode::eventOccurred(const ygString& event,const ygString& args)
    {
    //respond to event
    ygEvent newEvent(event,args);
    p_->eventMessages.respondToEvent(&newEvent);
    }


void ygNode::clearEventMessages(void)
    {
    p_->eventMessages.clear();
    }


void ygNode::clearEventMessages(const ygString& event)
    {
    p_->eventMessages.clear(event);
    }


void ygNode::clearEventMessages(const ygString& event,int num)
    {
    p_->eventMessages.clear(event,num);
    }


const std::map<ygString,std::vector<ygString> >& ygNode::getEventMessages(void)
    {
    return p_->eventMessages.getMessages();
    }


const std::vector<ygString>& ygNode::getEvents(void)
    {
    return p_->eventMessages.getEvents();
    }


const ygString& ygNode::getNextEvent(void)
    {
    return p_->eventMessages.getNextEvent();
    }


/***** Network functions *****/


void ygNode::unreliableMessage(const ygString& key)
    {
    p_->netMessages.unreliableMessage(key);
    }


void ygNode::reliableMessage(const ygString& key)
    {
    p_->netMessages.reliableMessage(key);
    }


void ygNode::addNetMessage(const ygString& message, void* data, int datatype)
    {
    p_->netMessages.addMessage(message,data,datatype);
    if (className() != netClassName())
        p_->netMessages.dontBroadcast(message);
    }


void ygNode::addNetMessageMap(const ygString& message, const ygString& token, int result)
    {
    p_->netMessages.addMessageMap(message,token,result);
    }


void ygNode::resetNetMessage(const ygString& message)
    {
    p_->netMessages.resetMessage(message);
    }


void ygNode::sendNetMessage(const ygMessage& msg)
    {
    msg.writeString(p_->netMessage);
    netMessageChanged("message",true);
    if (p_->debugNetMessages)
        cout << name() << " sending net message '" << msg << "'\n";
    }


void ygNode::sendNetMessage(const ygString& msgstr)
    {
    p_->netMessage = msgstr;
    netMessageChanged("message",true);
    if (p_->debugNetMessages)
        cout << name() << " sending net message '" << msgstr << "'\n";
    }


void ygNode::receiveNetMessage(const ygString& name,void* data)
    {
    if (p_->netMessages.receiveMessage(name,data))
        {
        ygMessage msg(name);
        message(msg);
        }
    }


ygString* ygNode::getNetMessageArg(const ygString& str)
    {
    char* s = str.c_str();
    int i;
    for (i=str.length()-1; i>0; i--)
        {
        if (!isdigit(s[i]))
        break;
        }
    ygString number;
    number.assign(str,i+1,str.length()-i-1);
    ygString message;
    message.assign(str,0,i+1);
    int index = atoi(number.c_str());
    return p_->netMessages.getMessageArg(message,index);
    }


void ygNode::parseMessage(const ygMessage& msg,bool force)
    {
    bool updated = false;
    updated = parseNetMessage(msg,force);
    message(msg);
    if (updated)
        {
        //the node has been updated by a message
        eventOccurred("updated");
        }
    }


bool ygNode::parseNetMessage(const ygMessage& msg,bool force)
    {
    if (p_->debugMessages)
        {
        ygString tmp;
        msg.writeString(tmp,false);
        cout << name() << " received message '" << tmp << "'\n";
        }
    if (isNetMaster() || force)
        return(p_->netMessages.parseMessage(msg));
    return false;
    }


void ygNode::netMessageChanged(const ygString& message,bool force)
    {
    if (isNetMaster() || force)
        p_->netMessages.sendMessage(message);
    }


void ygNode::dontSendNetMessage(const ygString& message)
    {
    p_->netMessages.dontBroadcast(message);
    }


void ygNode::dontReceiveNetMessage(const ygString& message)
    {
    p_->netMessages.dontReceive(message);
    }


void ygNode::dontSaveNetMessage(const ygString& message)
    {
    p_->netMessages.dontSave(message);
    }


void ygNode::requestNetMessages(void) const
    {
    p_->netMessages.requestAllMessages();
    }


bool ygNode::isNetMaster(void) const
    {
    return p_->isMaster;
    }


void ygNode::addComment(ygString* comment)
    {
    p_->comments.push_back(comment);
    }

const vector<ygString*>& ygNode::getComments(void)
    {
    return p_->comments;
    }

/***** Debug functions *****/


void ygNode::debugFlag(const ygString& flagname, bool* var)
    {
    p_->debug.addFlag(flagname,var);
    p_->debug.checkFlag(name(),flagname);
    p_->debug.checkFlag(className(),flagname);
    }


void ygNode::setDebug(const ygString& flag,bool val)
    {
    p_->debug.setFlag(flag,val);
    updateDebug();
    }


bool* ygNode::getDebug(const ygString& flag) const
    {
    return p_->debug.getFlag(flag);
    }


bool ygNode::debugDisplay(void) const
    {
    return p_->debugDisplay;
    }


bool ygNode::debugSelected(void) const
    {
    return p_->debugSelected;
    }


pfGroup* ygNode::debugNode(void) const
    {
    return p_->debugNode;
    }


pfVec4* ygNode::debugColor(void)
    {
    return p_->debugColor;
    }


pfGroup* ygNode::debugText(void)
    {
    if (!p_->debugText)
        createDebugText();
    return p_->debugText;
    }


pfString* ygNode::debugString(void)
    {
    return p_->debugString;
    }


void ygNode::setDebugNode(pfGroup* node)
    {
    node->setName(name().c_str());
    pfnode()->addChild(node);
    p_->debugNode = node;
    }


void ygNode::updateDebug(void)
    {
    if (!p_->debugNode)
        createDebugNode();
    if (p_->debugSelected)
        {
        p_->debugColor->set(YG_SELECT_COLOR);
        if (p_->debugString)
            p_->debugString->setColor(YG_SELECT_COLOR);
        }
    else
        {
        p_->debugColor->set(YG_DEBUG_COLOR);
        if (p_->debugString)
            p_->debugString->setColor(YG_DEBUG_COLOR);
        }
    if (p_->debugDisplay)
	{
        p_->debugNode->setTravMask(PFTRAV_DRAW, 0xffffffff, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
        p_->debugNode->setTravMask(PFTRAV_ISECT, YG_ISECT_ALLMASK|YG_ISECT_DEBUGMASK, PFTRAV_SELF|PFTRAV_DESCEND|PFTRAV_IS_CACHE, PF_SET);
	}
    else
	{
        p_->debugNode->setTravMask(PFTRAV_DRAW, 0, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
        p_->debugNode->setTravMask(PFTRAV_ISECT, 0, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
	}
    }


void ygNode::createDebugNode(void)
    {
    pfGeode* geode = new pfGeode;
    if (!geode)
        {
        cerr << "ERROR: ygNode::createDebugNode: failed to create pfGeode\n";
        return;
        }
    geode->setName(name().c_str());
    pfGeoSet* gset = new pfGeoSet;
    int* lengths = (int*) pfMalloc(3 * sizeof(int),pfGetSharedArena());
    gset->setPrimType(PFGS_POLYS);
    gset->setNumPrims(3);
    lengths[0] = 4;
    lengths[1] = 4;
    lengths[2] = 4;
    gset->setPrimLengths(lengths);
    pfVec3* verts = (pfVec3*) pfMalloc(12 * sizeof(pfVec3),pfGetSharedArena());
    verts[0][0] = -0.5; verts[0][1] = -0.5; verts[0][2] = 0;
    verts[1][0] = 0.5; verts[1][1] = -0.5; verts[1][2] = 0;
    verts[2][0] = 0.5; verts[2][1] = 0.5; verts[2][2] = 0;
    verts[3][0] = -0.5; verts[3][1] = 0.5; verts[3][2] = 0;
    verts[4][0] = 0; verts[4][1] = -0.5; verts[4][2] = -0.5;
    verts[5][0] = 0; verts[5][1] = 0.5; verts[5][2] = -0.5;
    verts[6][0] = 0; verts[6][1] = 0.5; verts[6][2] = 0.5;
    verts[7][0] = 0; verts[7][1] = -0.5; verts[7][2] = 0.5;
    verts[8][0] = -0.5; verts[8][1] = 0; verts[8][2] = -0.5;
    verts[9][0] = 0.5; verts[9][1] = 0; verts[9][2] = -0.5;
    verts[10][0] = 0.5; verts[10][1] = 0; verts[10][2] = 0.5;
    verts[11][0] = -0.5; verts[11][1] = 0; verts[11][2] = 0.5;
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);
    gset->setAttr(PFGS_COLOR4, PFGS_OVERALL, p_->debugColor, NULL);
    gset->setGState(new pfGeoState);
    gset->setLineWidth(1.0);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    gset->getGState()->setMode(PFSTATE_ENWIREFRAME,1);
    geode->addGSet(gset);
    pfnode()->addChild(geode);
    p_->debugNode = (pfGroup*)geode;
    }


void ygNode::createDebugText(void)
    {
    p_->debugString = new pfString;
    ygString fontFile("/usr/share/Performer/data");
    if (getenv("YG_PF_FONT_DIR"))
        fontFile = getenv("YG_PF_FONT_DIR");
    fontFile += "/Helvetica-Narrow";
    pfFont* font = pfdLoadFont_type1(fontFile.c_str(),PFDFONT_FILLED);
    p_->debugString->setFont(font);
    p_->debugString->setColor(YG_DEBUG_COLOR);
    p_->debugString->setMode(PFSTR_JUSTIFY,PFSTR_CENTER);
    p_->debugString->setString(name().c_str());
    pfBox box;
    float offset = (p_->debugString->getBBox()->max[2] + p_->debugString->getBBox()->min[2])/2;
    pfMatrix mat;
    mat.makeTrans(0,0,-offset);
    mat.postScale(mat,0.75,0.75,0.75);
    mat.postRot(mat,90.0,-1,0,0);
    p_->debugString->setMat(mat);
    float width = p_->debugString->getBBox()->max[0];
    pfGeoState* gstate = new pfGeoState;
    gstate->setMode(PFSTATE_ENLIGHTING,PF_OFF);
    pfGeoSet* chr = (pfGeoSet*)p_->debugString->getCharGSet(0);
    chr->setGState(gstate); 
    p_->debugString->setGState(gstate);
    pfText* text = new pfText;
    //set draw mask
    text->setTravMask(PFTRAV_DRAW,0xffffffff,PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
    //set intersection traversal mask
    text->setTravMask(PFTRAV_ISECT, YG_ISECT_ALLMASK | YG_ISECT_DEBUGMASK, PFTRAV_SELF|PFTRAV_DESCEND|PFTRAV_IS_CACHE, PF_SET);
    text->addString(p_->debugString);
    p_->debugText = new pfGroup;
    p_->debugText->addChild(text);
    pfGeode* geode = new pfGeode;
    geode->setName(name().c_str());
    pfGeoSet* gset = new pfGeoSet;
    int* lengths = (int*) pfMalloc(sizeof(int),pfGetSharedArena());
    gset->setPrimType(PFGS_POLYS);
    gset->setNumPrims(1);
    lengths[0] = 4;
    gset->setPrimLengths(lengths);
    pfVec3* verts = (pfVec3*) pfMalloc(4 * sizeof(pfVec3),pfGetSharedArena());
    verts[0][0] = -width-YG_DEBUG_PADDING; verts[0][1] = -0.2-YG_DEBUG_PADDING; verts[0][2] = 0.0;
    verts[1][0] = width+YG_DEBUG_PADDING; verts[1][1] = -0.2-YG_DEBUG_PADDING; verts[1][2] = 0.0;
    verts[2][0] = width+YG_DEBUG_PADDING; verts[2][1] = 0.2+YG_DEBUG_PADDING; verts[2][2] = 0.0;
    verts[3][0] = -width-YG_DEBUG_PADDING; verts[3][1] = 0.2+YG_DEBUG_PADDING; verts[3][2] = 0.0;
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);
    gset->setAttr(PFGS_COLOR4, PFGS_OVERALL, p_->debugColor, NULL);
    gset->setGState(new pfGeoState);
    gset->setLineWidth(1.0);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    gset->getGState()->setMode(PFSTATE_ENWIREFRAME,1);
    geode->addGSet(gset);  
    p_->debugText->addChild(geode);
    }


/***** Timing functions *****/


void ygNode::timingBegin(void)
    {
    p_->frameStart = ygGetTimeD();
    }


void ygNode::timingAppDone(void)
    {
    p_->appTimeTotal += ygGetTimeD() - p_->frameStart;
    }


void ygNode::timingEnd(void)
    {
    p_->frameTimeTotal += ygGetTimeD() - p_->frameStart;
    p_->frameCount++;
    }


/***** Print functions *****/


void ygNode::printMessages(ostream& s,int indent,bool comment)
    {
    if (comment)
        {
        vector<ygString*>::const_iterator iter;
        for (iter=p_->comments.begin(); iter != p_->comments.end();++iter)
            {
            for (int i=0; i < indent; i++)
                s << "    ";
            s << "//" << *(*iter);
            }
        }
    int padIndent = 0;
    padIndent = indent*4+className().length()+name().length()+2;
    for (int i=0; i < indent; i++)
        s << "    ";
    s << className() << " " << name() << "(";
    const vector<ygMessage*>& messages = p_->netMessages.getMessages(false);
    for (int i=0;i < messages.size(); i++)
        {
        if (i > 0)
            {
            s << "," << endl;
            for (int j=0; j < padIndent; j++)
                s << " ";
            }
        s << *(messages[i]);
        }
    bool found=false;
    const map<ygString,vector<ygString> >& events = p_->eventMessages.getMessages();
    map<ygString,vector<ygString> >::const_iterator i;
    for (i=events.begin(); i != events.end(); ++i)
        {
        const vector<ygString>& message = (*i).second;
        if (message.size() > 0)
            {
            if (found || messages.size() > 0)
                {
                s << "," << endl;
                for (int k=0; k < padIndent; k++)
                    s << " ";
                }
            found = true;
            for (int j=0;j < message.size(); j++)
                {
                if (j > 0)
                    {
                    s << "," << endl;
                    for (int k=0; k < padIndent; k++)
                        s << " ";
                    }
                s << "when(" << (*i).first << "," << message[j] << ")";
                }
            }
        }
    s << ")" << endl;
    }

void ygNode::printTimingStats(int indent)
    {
    int i, count;
    for (i=0; i < indent; i++)
        cout << "  ";
    count = p_->frameCount;
    if (count < 1)
        count = 1;
    cout << className() << " " << name() << " : "
        << p_->frameCount << " updates, "
        << p_->eventCount << " events, "
        << p_->frameTimeTotal*1000000.0/count << " usec/appTraverse, "
        << p_->appTimeTotal*1000000.0/count << " usec/app\n";
    for (i=0; i < numChildren(); i++)
        child(i)->printTimingStats(indent+1);
    p_->frameCount = 0;
    p_->eventCount = 0;
    p_->frameTimeTotal = 0;
    p_->appTimeTotal = 0;
    }
