// Description: load a texture and place it on a polygon
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the alias: texture
//<li> the initial polygon size is 1x1 oriented in the X-Y plane
//<li> image dimensions should be in powers of two: 32,64,128,256,512, or 1024
//<li> Currently the only format supported by ygTexture is SGI RGB format (*.rgb)
//<li> the gimp utility can be used to convert many formats to SGI RGB format in linux
//</ul>
//
// Category: Geometry
// Author: Alex Hill
//           11/01/01
//           10/15/02 Dave Pape - add optional cache flag to "file" msg, "reload" message, and included cache-state in netFilename
//           09/01/04 Alex Hill - derived from ygGeometry class and added functionality to create a parent ygTransform when necessary
// Revision: 06/28/05 Alex Hill - fixed several bugs related to performance with the material node
//
#include <Performer/pf/pfChannel.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfTexture.h>
#include <Performer/pr/pfMaterial.h>
#include <ygPFTextureCache.h>
#include "ygWorld.h"
#include "ygNetMessages.h"
#include "ygNodeDB.h"
#include "ygTexture.h"

extern "C" ygNode* construct_ygTexture(const char *name, bool master) { return new ygTexture(name, master); }

ygTexture::ygTexture(const char *name, bool master) : ygGeometry(name, master)
    {
    setClassName("ygTexture",true);
    gstate = NULL;
    geode = NULL;
    geoset = NULL;
    transform = NULL;
    cache = true;
    curCache = true;
    addNetMessage("file",&filename,YG_NET_FILE);
    addNetMessage("file",&curCache,YG_NET_BOOL);
    dontSendNetMessage("file");
    //create texture polygon
    createGeometry();
    }

ygTexture::~ygTexture(void)
    {
    if (gstate && curCache)
        gstate->setAttr(PFSTATE_TEXTURE, NULL);
    }

void ygTexture::reset(void)
    {
    //set cache to true
    cache = true;
    curCache = true;
    resetNetMessage("file");
    if (!transform && parent())
        {
        ygString transName(name());
        transName += "xform";
        if (parent()->name() == transName)
            transform = parent();
        }
    ygGeometry::reset();
    }


void ygTexture::message(const ygMessage& msg)
    {
    //load the given texture file with optional caching
    if (msg == "file")
        {
        if (filename.length() > 0)
            setTexture(filename);
        }
    //set the default object caching mode
    else if (msg == "cache")
        {
        if (msg.numArgs() > 0)
            cache = msg.boolArg(0);
        }
    //reload the currently loaded object
    else if (msg == "reload")
        {
        curCache = false;
        setTexture(filename);
        }
    //set whether the object is drawn locally
    else if (msg == "showLocal")
        {
        if (msg.numArgs() > 0)
            {
            bool val = msg.boolArg(0);
            if(val)
                setDraw(DRAW_ON);
            else
                setDraw(DRAW_REMOTE);
            }
        else
            setDraw(DRAW_ON);
        }
    //create a parent ygTransform and pass the message
    else if (msg == "position" || msg == "orientation" || msg == "size" || msg == "updateInterval" || msg == "reliableKey")
        {
        createTransform();
        transform->parseMessage(msg);
        }
    else
        ygGeometry::message(msg);
    }

void ygTexture::setTexture(const ygString& file)
    {
    pfTexture* texture;
    //if cache mode is true then retrieve from texture cache
    if (!cache)
        curCache = false;
    if (curCache)
        texture = ygPFTextureCache::textureReference(file.c_str());
    //else, load the file using loadFile
    else
        {
        texture = new pfTexture;
        texture->loadFile(file.c_str());
        }
    if (texture)
        {
        filename = file;
        gstate->setAttr(PFSTATE_TEXTURE, texture);
        gstate->setMode(PFSTATE_ENTEXTURE, PF_ON);
        int comp;
        //get number of image components
        texture->getImage(NULL,&comp,NULL,NULL,NULL);
        //if alpha channel then set transparency
        if (comp == 4)
            {
            gstate->setMode(PFSTATE_TRANSPARENCY,PFTR_BLEND_ALPHA);
            geoset->setDrawBin(PFSORT_TRANSP_BIN);
            }
        else
            {
            gstate->setMode(PFSTATE_TRANSPARENCY,PFTR_OFF);
            geoset->setDrawBin(PFSORT_OPAQUE_BIN);
            }
        geode->setTravMask(PFTRAV_DRAW,0xffffffff,PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
        //set intersection traversal mask
        geode->setTravMask(PFTRAV_ISECT, YG_ISECT_ALLMASK, PFTRAV_SELF|PFTRAV_DESCEND|PFTRAV_IS_CACHE, PF_SET);
        netMessageChanged("file");
        setChanged(YG_CHANGE_TEXTURE);
        }
    }

void ygTexture::createGeometry(void)
    {        
    //create a new pfGeode and add it to the scene
    geode = new pfGeode;
    
    //allocate (from shared memory) an array of vertices for the polygon
    pfVec3 *vertex = (pfVec3 *) pfMalloc(4*sizeof(pfVec3),pfGetSharedArena());
    vertex[0].set( 0.5, -0.5, 0.0);
    vertex[1].set( 0.5, 0.5, 0.0);
    vertex[2].set(-0.5, 0.5, 0.0);
    vertex[3].set(-0.5, -0.5, 0.0);

    //create an array of normals for the polygon
    pfVec3 *normal = (pfVec3 *) pfMalloc(4*sizeof(pfVec3),pfGetSharedArena());
    normal[0].set(0.0, 0.0, 1.0);
    normal[1].set(0.0, 0.0, 1.0);
    normal[2].set(0.0, 0.0, 1.0);
    normal[3].set(0.0, 0.0, 1.0);

    //create an array of texture coordinates for the polygon
    pfVec2 *tCoord = (pfVec2 *) pfMalloc(4*sizeof(pfVec2),pfGetSharedArena());
    tCoord[0].set(1.0, 0.0);
    tCoord[1].set(1.0, 1.0);
    tCoord[2].set(0.0, 1.0);
    tCoord[3].set(0.0, 0.0);

    //create the geometry set and add the geometry data to it
    geoset = new pfGeoSet;
    geoset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, vertex, NULL);
    geoset->setAttr(PFGS_NORMAL3, PFGS_PER_VERTEX, normal, NULL);
    geoset->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, tCoord, NULL);
    geoset->setPrimType(PFGS_QUADS);
    geoset->setNumPrims(1);
    geode->addGSet(geoset);
    
    //Create a new material
    pfMaterial *material = new pfMaterial;
    
    //Create a pfGeostate, add the material, and then assign it to our geometry
    gstate = new pfGeoState;
    gstate->setAttr(PFSTATE_FRONTMTL, material);
    gstate->setAttr(PFSTATE_BACKMTL,  material);
    gstate->setMode(PFSTATE_CULLFACE, PF_OFF);
    gstate->setAttr(PFSTATE_TEXENV, new pfTexEnv);
    geoset->setGState(gstate);
    geode->setName(name().c_str());
    object->addChild(geode);
    setChanged(YG_CHANGE_GEOMETRY);
    }
    
void ygTexture::createTransform(void)
    {
    if (!transform)
        {
        ygString transName(name());
        transName += "xform";
        ygNodeDB::createValidName("ygTransform",transName);
        transform = ygNodeDB::create("ygTransform",transName);
        parent()->addChild(transform);
        transform->addChild(this);
        transform->reset();
        }
    }

void ygTexture::createDebugNode(void)
    {
    createTransform();
    ygGeometry::createDebugNode();
    }
