// Description: creates a transformation matrix with position, orientation, and scale
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the aliases: transform, ygSimpleTransform, and simpleTransform
//<li> the orientation message is measured in degrees and allows you to control the heading, pitch, and roll of the child node
//<li> creates a pfDCS and maintains network state with a pfCoord for translation and rotation and a pfVec3 for size
//</ul>
//
// Category: Transformation
// Author: Dave Pape
//           11/01/01
//           09/01/04 Alex Hill - merged with ygSimpleTransform and added "coordinates" debug
//           06/28/05 Alex Hill - fixed orientation update problem in setMatrix
//           02/01/06 Alex Hill - added logic to prevent zero size matrices and added debug display code
// Revision: 07/01/06 Alex Hill - added logic to switch to unreliable messages when possible
//
#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfMaterial.h>
#include <Performer/pfdu.h>
#include "ygNetMessages.h"
#include "ygWorld.h"
#include "ygTransform.h"
#include "ygUtil.h"

extern "C" ygNode* construct_ygTransform(const char* name,bool master) { return new ygTransform(name,master); }

#define MIN_SIZE 0.0000001

#define CONE_BASE 0.0
#define CONE_TOP 0.1
#define CONE_RADIUS 0.1
#define CONE_LENGTH 0.070711 
#define AXIS_WIDTH 0.01

#define CHANGE_THRESHOLD 3
#define CHANGE_INITIALIZED -1

ygTransform::ygTransform(const char* name,bool master) : ygNode(name,master,false)
    {
    setClassName("ygTransform",true);
    //create new pfDCS and set as pfnode
    dcs = new pfDCS;
    setPfNode(dcs);
    netCoord.xyz.set(0,0,0);
    netCoord.hpr.set(0,0,0);
    lastCoord = netCoord;
    lastNetCoord = netCoord;
    netSize.set(1,1,1);
    ori.set(0,0,0);
    lastSize = netSize;
    lastNetSize = netSize;
    lastRadius = 0;
    netCoordAuto = true;
    netSizeAuto = true;
    netCoordUnchanged = CHANGE_INITIALIZED;
    netSizeUnchanged = CHANGE_INITIALIZED;
    updateInterval = 0;
    if (getenv("YG_NET_TRANSFORM_INTERVAL"))
        updateInterval = atof(getenv("YG_NET_TRANSFORM_INTERVAL"));
    lastNetUpdate = 0;
    addNetMessage("size",&netSize,YG_NET_VEC3);
    dontSendNetMessage("size");
    addNetMessage("position",&netCoord.xyz,YG_NET_VEC3);
    dontSendNetMessage("position");
    addNetMessage("orientation",&ori,YG_NET_VEC3);
    dontSendNetMessage("orientation");
    addNetMessage("updateInterval",&updateInterval,YG_NET_FLOAT);
    //create a simple unreliable coordinate network key
    addNetMessage(COORD_NETKEY, &netCoord, YG_NET_COORD);
    //dont include in documentation
    dontSaveNetMessage(COORD_NETKEY);
    reliableMessage(COORD_NETKEY);
#if 0
    addNetMessage(MATRIX_NETKEY, &netMatrix, YG_NET_MATRIX);
    //dont include in documentation
    dontSaveNetMessage(MATRIX_NETKEY);
    unreliableMessage(MATRIX_NETKEY);
#endif
    //register changed event
    addEvent("changed");
    }


ygTransform::~ygTransform(void)
    {
    }


void ygTransform::reset(void)
    {
    //if net master then set matrix to identity
    if (isNetMaster())
        {
        pfMatrix mat;
        mat.makeIdent();
        setMatrix(mat);
        //reset update interval
        resetNetMessage("updateInterval");
        //reset last update timer
        lastNetUpdate = 0;
        netCoordAuto = true;
        netSizeAuto = true;
        netCoordUnchanged = CHANGE_INITIALIZED;
        netSizeUnchanged = CHANGE_INITIALIZED;
        reliableMessage(COORD_NETKEY);
        reliableMessage("size");
        }
    ygNode::reset();
    }


/***** Message functions *****/

void ygTransform::message(const ygMessage& msg)
    {
    //set the translational component
    if (msg == "position")
        {
        setPosition(netCoord.xyz);
        }
    //set the rotational component applied before translation
    else if (msg == "orientation")
        {
        setOrientation(ori);
        }
    //set the scaling component
    else if (msg == "size")
        {
        if (msg.numArgs() == 1)
            {
            netSize[1] = netSize[0];
            netSize[2] = netSize[0];
            }
        if (netSize[0] < MIN_SIZE)
            netSize[0] = MIN_SIZE;
        if (netSize[1] < MIN_SIZE)
            netSize[1] = MIN_SIZE;
        if (netSize[2] < MIN_SIZE)
            netSize[2] = MIN_SIZE;
        setSize(netSize);
        }
    //set the time interval to update remote users
    else if (msg == "updateInterval")
        {    
        return;
        }
    else if (msg == COORD_NETKEY)
        {
        setPosition(netCoord.xyz);
        setOrientation(netCoord.hpr[1],netCoord.hpr[2],netCoord.hpr[0]);
        }
#if 0
    else if (msg == MATRIX_NETKEY)
        {
        setMatrix(netMatrix);
        }
#endif
    else
        {
        if ((msg == "unreliableKey" || msg == "reliablekey") && msg.numArgs() > 0)
            {
            ygString key = msg.stringArg(0);
            if (key == COORD_NETKEY)
                netCoordAuto = false;
            else if (key == "size")
                netSizeAuto = false;
            }
       ygNode::message(msg);
       }
    }


/*****  DCS functions *****/

void ygTransform::getMatrix(pfMatrix& mat) const
    {
    dcs->getMat(mat);
    }


void ygTransform::getPosition(pfVec3& vec) const
    {
    vec = netCoord.xyz;
    }


void ygTransform::getOrientation(pfVec3& vec) const
    {
    vec = ori;
    }


void ygTransform::getSize(pfVec3& vec) const
    {
    vec = netSize;
    }


void ygTransform::setMatrix(pfMatrix& mat)
    {
    dcs->setMat(mat);
    pfVec3d row;
    row.set(mat.mat[0][0],mat.mat[0][1],mat.mat[0][2]);
    netSize[0] = row.length();
    row.set(mat.mat[1][0],mat.mat[1][1],mat.mat[1][2]);
    netSize[1] = row.length();
    row.set(mat.mat[2][0],mat.mat[2][1],mat.mat[2][2]);
    netSize[2] = row.length();
    pfMatrix unscaled;
    unscaled.preScale(1.0/netSize[0],1.0/netSize[1],1.0/netSize[2],mat);
    unscaled.getOrthoCoord(&(netCoord));
    if (netCoord.hpr[0] == -180.0 && netCoord.hpr[2] == 180.0)
        {
        netCoord.hpr[0] = 0.0;
        if (netCoord.hpr[1] >= 0.0)
            netCoord.hpr[1] = 180.0 - netCoord.hpr[1];
        else
            netCoord.hpr[1] = -(180.0 + netCoord.hpr[1]);
        netCoord.hpr[2] = 0.0;
        }
    ori[0] = netCoord.hpr[1];
    ori[1] = netCoord.hpr[2];
    ori[2] = netCoord.hpr[0];
    }


void ygTransform::setPosition(float x,float y,float z)
    {
    dcs->setTrans(x,y,z);
    netCoord.xyz.set(x,y,z);
    }


void ygTransform::setOrientation(float x,float y,float z)
    {
    ori.set(x,y,z);
    dcs->setRot(z,x,y);
    netCoord.hpr.set(z,x,y);
    }


void ygTransform::setSize(float s)
    {
    netSize.set(s,s,s);
    dcs->setScale(s);
    }


void ygTransform::setSize(float x,float y,float z)
    {
    netSize.set(x,y,z);
    dcs->setScale(x,y,z);
    }


void ygTransform::app(void)
    {
    bool sizeChange = false;
    if (netSize != lastSize)
        {
        sizeChange = true;
        lastSize = netSize;
        }
    //adjust debug display
    if (debugDisplay())
        {
        pfMatrix mat;
        getMatrix(mat);
        pfVec3 pos(0,0,0);
        pos.xformPt(pos,mat);
        pfVec3 vec;
        float size=0.0;
        if (netSize[0] > netSize[1] && netSize[0] > netSize[2])
            size = netSize[0];
        else if (netSize[1] > netSize[2])
            size = netSize[1];
        else
            size = netSize[2];
        pfSphere sphere;
        sphere.radius = 0;
        sphere.center.set(0,0,0);
        float radius = 0;
        pfNode* debug = debugNode();
        for (int i=0; i<pfnode()->getNumChildren(); i++)
            {
            pfNode* child = pfnode()->getChild(i);
            if (child != debug)
                {
                child->getBound(&sphere);
                sphere.center.xformPt(sphere.center,mat);
                sphere.radius *= size;
                vec = sphere.center - pos;
                float currRadius = vec.length() + sphere.radius;
                if (currRadius > radius)
                    radius = currRadius;
                }
            }
        if (radius == 0.0)
            radius = size/2;
        //if bounding sphere or size has changed then update debug
        if (sizeChange || radius != lastRadius)
            {
            setDebugCoords(radius);
            lastRadius = radius;
            }
        }
    //if net master
    if (isNetMaster())
        {
        //if update interval has passed then
        if (updateInterval >= 0.0 && ygWorld::frameTime()-lastNetUpdate > updateInterval)
            {
            //if network coord has changed then
            if (netCoord.xyz != lastNetCoord.xyz ||
                netCoord.hpr != lastNetCoord.hpr)
                {
                //indicate message changed
                netMessageChanged(COORD_NETKEY);
                lastNetCoord = netCoord;
                lastNetUpdate = ygWorld::frameTime();
                //if unchanged counter is positive then
                if (netCoordUnchanged >= 0)
                    {
                    //if counter is over threshold then set unreliable
                    if (netCoordAuto && netCoordUnchanged > CHANGE_THRESHOLD)
                        unreliableMessage(COORD_NETKEY);
                    //set counter to zero
                    netCoordUnchanged = 0;
                    }
                //else initialize counter over threshold
                else 
                    netCoordUnchanged = CHANGE_THRESHOLD+1;
                }
            //else if counter is under threshold then increment
            else if (netCoordUnchanged < CHANGE_THRESHOLD)
                {
                netCoordUnchanged++;
                netMessageChanged(COORD_NETKEY);
                }
            //else if counter is at threshold then set reliable
            else if (netCoordUnchanged == CHANGE_THRESHOLD)
                {
                netCoordUnchanged++;
                if (netCoordAuto)
                    reliableMessage(COORD_NETKEY);
                }
            //if network size has changed then
            if (netSize != lastNetSize)
                {
                //indicate message changed
                netMessageChanged("size");
                lastNetSize = netSize;
                lastNetUpdate = ygWorld::frameTime();
                //if unchanged counter is positive then
                if (netSizeUnchanged >= 0)
                    {
                    //if counter is over threshold then set unreliable
                    if (netSizeAuto && netSizeUnchanged > CHANGE_THRESHOLD)
                        unreliableMessage("size");
                    //set counter to zero
                    netSizeUnchanged = 0;
                    }
                //else initialize counter over threshold
                else 
                    netSizeUnchanged = CHANGE_THRESHOLD+1;
                }
            //else if counter is under threshold then increment
            else if (netSizeUnchanged < CHANGE_THRESHOLD)
                netSizeUnchanged++;
            //else if counter is at threshold then set reliable
            else if (netSizeUnchanged == CHANGE_THRESHOLD)
                {
                netSizeUnchanged++;
                if (netSizeAuto)
                    reliableMessage("size");
                netMessageChanged("size");
                }
            }
         if (sizeChange || 
             netCoord.xyz != lastCoord.xyz ||
             netCoord.hpr != lastCoord.hpr)
                {
                //the transform has changed
                eventOccurred("changed");
                lastCoord = netCoord;
                }
        }
    ygNode::app();
    }


void ygTransform::setTransform(const pfMatrix& mat)
    {
    pfMatrix parentMat;
    getParentTransform(parentMat);
    pfMatrix inverse;
    inverse.invertAff(parentMat);
    pfMatrix newMat;
    newMat = mat*inverse;
    setMatrix(newMat);
    }


void ygTransform::createDebugNode(void)
    {
    pfGeode* geode = new pfGeode;
    if (!geode)
        {
        cerr << "ERROR: ygTransform::createDebugGeode: failed to create pfGeode\n";
        return;
        }
    //create cone geoset
    pfGeoSet *gset = new pfGeoSet;
    int* lengths = (int *) pfMalloc(6*sizeof(int),pfGetSharedArena());
    gset->setPrimType(PFGS_TRISTRIPS);
    gset->setNumPrims(6);
    lengths[0] = 18;
    lengths[1] = 18;
    lengths[2] = 18;
    lengths[3] = 18;
    lengths[4] = 18;
    lengths[5] = 18;
    gset->setPrimLengths(lengths);
    debugVerts = (pfVec3*)pfMalloc(132*sizeof(pfVec3),pfGetSharedArena());
    for(int i=0; i<132; i++)
        {
        debugVerts[i].vec[0] = 0.0;
        debugVerts[i].vec[1] = 0.0;
        debugVerts[i].vec[2] = 0.0;
        }
    gset->setAttr(PFGS_COORD3,PFGS_PER_VERTEX,debugVerts,NULL);
    pfVec4 *colorAttr = (pfVec4*)pfMalloc(6*sizeof(pfVec4),pfGetSharedArena());
    colorAttr[0].set(1,0,0,1);
    colorAttr[1].set(1,0,0,1);
    colorAttr[2].set(0,1,0,1);
    colorAttr[3].set(0,1,0,1);
    colorAttr[4].set(0,0,1,1);
    colorAttr[5].set(0,0,1,1);
    gset->setAttr(PFGS_COLOR4,PFGS_PER_PRIM,colorAttr,NULL);
    gset->setLineWidth(0.5);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    gset->setBound(NULL,PFBOUND_DYNAMIC);
    pfGeoState* gstate = new pfGeoState;
    gstate->setMode(PFSTATE_ENWIREFRAME,PF_ON);
    gset->setGState(gstate);
    geode->addGSet(gset);
    //create axis geoset
    gset = new pfGeoSet;
    lengths = (int*)pfMalloc(6*sizeof(int),pfGetSharedArena());
    gset->setPrimType(PFGS_POLYS);
    gset->setNumPrims(6);
    lengths[0] = 4;
    lengths[1] = 4;
    lengths[2] = 4;
    lengths[3] = 4;
    lengths[4] = 4;
    lengths[5] = 4;
    gset->setPrimLengths(lengths);
    gset->setAttr(PFGS_COORD3,PFGS_PER_VERTEX,debugVerts+6*18,NULL);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    gset->setLineWidth(1.0);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    gset->setBound(NULL,PFBOUND_DYNAMIC);
    gset->setGState(gstate);
    geode->addGSet(gset);
    setDebugNode((pfGroup*)geode);
    setDebugCoords(0.5);
    }


void ygTransform::setDebugCoords(float radius)
    {
    // CONE_LENGTH = sqrt(CONE_RADIUS^2 + CONE_RADIUS^2)/2
    pfVec3* verts = debugVerts;
    float x_base = (radius + CONE_BASE)/netSize.vec[0];
    float y_base = (radius + CONE_BASE)/netSize.vec[1];
    float z_base = (radius + CONE_BASE)/netSize.vec[2];
    float x_top = (radius + CONE_TOP)/netSize.vec[0];
    float y_top = (radius + CONE_TOP)/netSize.vec[1];
    float z_top = (radius + CONE_TOP)/netSize.vec[2];
    float x_radius = CONE_RADIUS/netSize.vec[0];
    float y_radius = CONE_RADIUS/netSize.vec[1];
    float z_radius = CONE_RADIUS/netSize.vec[2];
    float x_length = CONE_LENGTH/netSize.vec[0];
    float y_length = CONE_LENGTH/netSize.vec[1];
    float z_length = CONE_LENGTH/netSize.vec[2];
    float x_width = AXIS_WIDTH/netSize.vec[0];
    float y_width = AXIS_WIDTH/netSize.vec[1];
    float z_width = AXIS_WIDTH/netSize.vec[2];
    //cone vertices
    verts[0].vec[0] = x_base; verts[0].vec[1] = y_radius;
    verts[1].vec[0] = x_top;
    verts[2].vec[0] = x_base; verts[2].vec[1] = y_length; verts[2].vec[2] = -z_length;
    verts[3].vec[0] = x_top;
    verts[4].vec[0] = x_base; verts[4].vec[2] = -z_radius;
    verts[5].vec[0] = x_top;
    verts[6].vec[0] = x_base; verts[6].vec[1] = -y_length; verts[6].vec[2] = -z_length;
    verts[7].vec[0] = x_top;
    verts[8].vec[0] = x_base; verts[8].vec[1] = -y_radius;
    verts[9].vec[0] = x_top;
    verts[10].vec[0] = x_base; verts[10].vec[1] = -y_length; verts[10].vec[2] = z_length;
    verts[11].vec[0] = x_top;
    verts[12].vec[0] = x_base; verts[12].vec[2] = z_radius;
    verts[13].vec[0] = x_top;
    verts[14].vec[0] = x_base; verts[14].vec[1] = y_length; verts[14].vec[2] = z_length;
    verts[15].vec[0] = x_top;
    verts[16].vec[0] = x_base; verts[16].vec[1] = y_radius;
    verts[17].vec[0] = x_top;
    verts[18].vec[0] = -x_base; verts[18].vec[1] = y_radius;
    verts[19].vec[0] = -x_top;
    verts[20].vec[0] = -x_base; verts[20].vec[1] = y_length; verts[20].vec[2] = z_length;
    verts[21].vec[0] = -x_top;
    verts[22].vec[0] = -x_base; verts[22].vec[2] = z_radius;
    verts[23].vec[0] = -x_top;
    verts[24].vec[0] = -x_base; verts[24].vec[1] = -y_length; verts[24].vec[2] = z_length;
    verts[25].vec[0] = -x_top;
    verts[26].vec[0] = -x_base; verts[26].vec[1] = -y_radius;
    verts[27].vec[0] = -x_top;
    verts[28].vec[0] = -x_base; verts[28].vec[1] = -y_length; verts[28].vec[2] = -z_length;
    verts[29].vec[0] = -x_top;
    verts[30].vec[0] = -x_base; verts[30].vec[2] = -z_radius;
    verts[31].vec[0] = -x_top;
    verts[32].vec[0] = -x_base; verts[32].vec[1] = y_length; verts[32].vec[2] = -z_length;
    verts[33].vec[0] = -x_top;
    verts[34].vec[0] = -x_base; verts[34].vec[1] = y_radius;
    verts[35].vec[0] = -x_top;
    verts[36].vec[1] = y_base; verts[36].vec[2] = -z_radius;
    verts[37].vec[1] = y_top;
    verts[38].vec[0] = x_length; verts[38].vec[1] = y_base; verts[38].vec[2] = -z_length;
    verts[39].vec[1] = y_top;
    verts[40].vec[0] = x_radius; verts[40].vec[1] = y_base;
    verts[41].vec[1] = y_top;
    verts[42].vec[0] = x_length; verts[42].vec[1] = y_base; verts[42].vec[2] = z_length;
    verts[43].vec[1] = y_top;
    verts[44].vec[1] = y_base; verts[44].vec[2] = z_radius;
    verts[45].vec[1] = y_top;
    verts[46].vec[0] = -x_length; verts[46].vec[1] = y_base; verts[46].vec[2] = z_length;
    verts[47].vec[1] = y_top;
    verts[48].vec[0] = -x_radius; verts[48].vec[1] = y_base;
    verts[49].vec[1] = y_top;
    verts[50].vec[0] = -x_length; verts[50].vec[1] = y_base; verts[50].vec[2] = -z_length;
    verts[51].vec[1] = y_top;
    verts[52].vec[1] = y_base; verts[52].vec[2] = -z_radius;
    verts[53].vec[1] = y_top;
    verts[54].vec[1] = -y_base; verts[54].vec[2] = z_radius;
    verts[55].vec[1] = -y_top;
    verts[56].vec[0] = x_length; verts[56].vec[1] = -y_base; verts[56].vec[2] = z_length;
    verts[57].vec[1] = -y_top;
    verts[58].vec[0] = x_radius; verts[58].vec[1] = -y_base;
    verts[59].vec[1] = -y_top;
    verts[60].vec[0] = x_length; verts[60].vec[1] = -y_base; verts[60].vec[2] = -z_length;
    verts[61].vec[1] = -y_top;
    verts[62].vec[1] = -y_base; verts[62].vec[2] = -z_radius;
    verts[63].vec[1] = -y_top;
    verts[64].vec[0] = -x_length; verts[64].vec[1] = -y_base; verts[64].vec[2] = -z_length;
    verts[65].vec[1] = -y_top;
    verts[66].vec[0] = -x_radius; verts[66].vec[1] = -y_base;
    verts[67].vec[1] = -y_top;
    verts[68].vec[0] = -x_length; verts[68].vec[1] = -y_base; verts[68].vec[2] = z_length;
    verts[69].vec[1] = -y_top;
    verts[70].vec[1] = -y_base; verts[70].vec[2] = z_radius;
    verts[71].vec[1] = -y_top;
    verts[72].vec[1] = y_radius; verts[72].vec[2] = z_base;
    verts[73].vec[2] = z_top;
    verts[74].vec[0] = x_length; verts[74].vec[1] = y_length; verts[74].vec[2] = z_base;
    verts[75].vec[2] = z_top;
    verts[76].vec[0] = x_radius; verts[76].vec[2] = z_base;
    verts[77].vec[2] = z_top;
    verts[78].vec[0] = x_length; verts[78].vec[1] = -y_length; verts[78].vec[2] = z_base;
    verts[79].vec[2] = z_top;
    verts[80].vec[1] = -y_radius; verts[80].vec[2] = z_base;
    verts[81].vec[2] = z_top;
    verts[82].vec[0] = -x_length; verts[82].vec[1] = -y_length; verts[82].vec[2] = z_base;
    verts[83].vec[2] = z_top;
    verts[84].vec[0] = -x_radius; verts[84].vec[2] = z_base;
    verts[85].vec[2] = z_top;
    verts[86].vec[0] = -x_length; verts[86].vec[1] = y_length; verts[86].vec[2] = z_base;
    verts[87].vec[2] = z_top;
    verts[88].vec[1] = y_radius; verts[88].vec[2] = z_base;
    verts[89].vec[2] = z_top;
    verts[90].vec[1] = -y_radius; verts[90].vec[2] = -z_base;
    verts[91].vec[2] = -z_top;
    verts[92].vec[0] = x_length; verts[92].vec[1] = -y_length; verts[92].vec[2] = -z_base;
    verts[93].vec[2] = -z_top;
    verts[94].vec[0] = x_radius; verts[94].vec[2] = -z_base;
    verts[95].vec[2] = -z_top;
    verts[96].vec[0] = x_length; verts[96].vec[1] = y_length; verts[96].vec[2] = -z_base;
    verts[97].vec[2] = -z_top;
    verts[98].vec[1] = y_radius; verts[98].vec[2] = -z_base;
    verts[99].vec[2] = -z_top;
    verts[100].vec[0] = -x_length; verts[100].vec[1] = y_length; verts[100].vec[2] = -z_base;
    verts[101].vec[2] = -z_top;
    verts[102].vec[0] = -x_radius; verts[102].vec[2] = -z_base;
    verts[103].vec[2] = -z_top;
    verts[104].vec[0] = -x_length; verts[104].vec[1] = -y_length; verts[104].vec[2] = -z_base;
    verts[105].vec[2] = -z_top;
    verts[106].vec[1] = -y_radius; verts[106].vec[2] = -z_base;
    verts[107].vec[2] = -z_top;
    //axis vertices
    verts[108].vec[0] = -x_base; verts[108].vec[1] = -y_width;
    verts[109].vec[0] = x_base; verts[109].vec[1] = -y_width;
    verts[110].vec[0] = x_base; verts[110].vec[1] = y_width;
    verts[111].vec[0] = -x_base; verts[111].vec[1] = y_width;
    verts[112].vec[0] = -x_base; verts[112].vec[2] = -z_width;
    verts[113].vec[0] = x_base; verts[113].vec[2] = -z_width;
    verts[114].vec[0] = x_base; verts[114].vec[2] = z_width;
    verts[115].vec[0] = -x_base; verts[115].vec[2] = z_width;
    verts[116].vec[1] = -y_base; verts[116].vec[2] = -z_width;
    verts[117].vec[1] = y_base; verts[117].vec[2] = -z_width;
    verts[118].vec[1] = y_base; verts[118].vec[2] = z_width;
    verts[119].vec[1] = -y_base; verts[119].vec[2] = z_width;
    verts[120].vec[0] = -x_width; verts[120].vec[1] = -y_base;
    verts[121].vec[0] = -x_width; verts[121].vec[1] = y_base;
    verts[122].vec[0] = x_width; verts[122].vec[1] = y_base;
    verts[123].vec[0] = x_width; verts[123].vec[1] = -y_base;
    verts[124].vec[0] = -x_width; verts[124].vec[2] = -z_base;
    verts[125].vec[0] = -x_width; verts[125].vec[2] = z_base;
    verts[126].vec[0] = x_width; verts[126].vec[2] = z_base;
    verts[127].vec[0] = x_width; verts[127].vec[2] = -z_base;
    verts[128].vec[1] = -y_width; verts[128].vec[2] = -z_base;
    verts[129].vec[1] = -y_width; verts[129].vec[2] = z_base;
    verts[130].vec[1] = y_width; verts[130].vec[2] = z_base;
    verts[131].vec[1] = y_width; verts[131].vec[2] = -z_base;
    pfBox box;
    box.min.set(-x_top,-y_top,-z_top);
    box.max.set(x_top,y_top,z_top);
    ((pfGeode*)debugNode())->getGSet(0)->setBound(&box,PFBOUND_STATIC);
    ((pfGeode*)debugNode())->getGSet(1)->setBound(&box,PFBOUND_STATIC);
    debugNode()->setTravMask(PFTRAV_ISECT, YG_ISECT_ALLMASK|YG_ISECT_DEBUGMASK, PFTRAV_SELF|PFTRAV_DESCEND|PFTRAV_IS_CACHE, PF_SET);
    pfSphere sphere;
    if (netSize[0] < netSize[1] && netSize[0] < netSize[2])
        sphere.radius = (radius + CONE_TOP)/netSize[0];
    else if (netSize[1] < netSize[2])
        sphere.radius = (radius + CONE_TOP)/netSize[1];
    else
        sphere.radius = (radius + CONE_TOP)/netSize[2];
    sphere.center.set(0,0,0);
    debugNode()->setBound(NULL,PFBOUND_DYNAMIC);
    }

