// Description: the class that loads scene files and maintains the scene graph
//
//<b>notes:</b>
//<ul>
//<li> this node establishes the local world for each client session
//<li> messages can be sent to the ygWorld instance by addressing them to "world" or "localroot" (i.e. world.quit)
//<li> adding world.sentmessages to the YG_DEBUG variable will print out each message that is sent
//</ul>
//
// Category: Foundation
// Author: Dave Pape
//           11/01/01
//           10/01/02 Alex Hill - added pfChanged and setPFChanged
//           08/01/03 Alex Hill - changed ygSound references to use ygSoundServer
//           09/01/04 Alex Hill - added create, delete, and debug messages
//           06/28/05 Alex Hill - adjusted timing calculation to compenstate for long load times
//           02/01/06 Alex Hill - changed derivation to ygNode and added frame count message delay queue
// Revision: 07/01/06 Alex Hill - fixed bug initializing timing
//
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <list>
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfEarthSky.h>
#include <Performer/pr/pfFog.h>
#include <Performer/pf/pfScene.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfGeode.h>
#include <pfcave.h>
#include "ygParser.h"
#include "ygUtil.h"
#include "ygNodeDB.h"
#include "ygNet.h"
#include "ygNetMessages.h"
#include "ygDebugFlags.h"
#include "ygSoundServer.h"
#include "ygCAVEViewer.h"
#include "ygWorld.h"

using namespace std;

class ygDelayMsg
    {
    public:
     ygMessage msg;
     float time;
     int frame;
    };

struct _ygWorldPrivateData
    {
    bool sentMessageDebug;
    list<ygDelayMsg*> messageQueue;
    list<ygDelayMsg*> messageFrameQueue;
    pfScene* pfscene;
    pfEarthSky* pfsky;
    pfFog* pffog;
    vector<ygUser*> users;
    ygNode* viewer;
    ygNode* root_;
    ygString saveFile;
    ygString saveFrom;
    ygString editType;
    ygString editFrom;
    ygString editTypeLast;
    ygString editFromLast;
    ygString loadFile;
    ygString loadFrom;
    ygString deleteFrom;
    bool retainRoot;
    float lastTime;
    bool quitFlag;
    };

float ygWorld::frameTime_ = 0;
float ygWorld::frameDeltaTime_ = 1.0;
int ygWorld::frameNumber_ = 0;
ygWorld* ygWorld::world_ = NULL;


extern "C" ygNode* construct_ygWorld(const char* name,bool master) { return new ygWorld(name,master); }

ygWorld::ygWorld(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("ygWorld");
    if (world_)
        cerr << "ERROR: More than one ygWorld has been created -"
            " there should be only one!\n";
    world_ = this;
    p_ = new struct _ygWorldPrivateData;
    p_->pfscene = new pfScene;
    p_->pfsky = new pfEarthSky;
    p_->pffog = new pfFog;
    pfCAVEMasterChan()->setESky(p_->pfsky);
    debugFlag("SentMessages",&p_->sentMessageDebug);
    p_->sentMessageDebug = ygDebugFlags::checkDebugEnv("world.sentmessages");
    ygString viewerName(name);
    viewerName += "_viewer";
    ygNodeDB::createValidName("ygCAVEViewer",viewerName);
    p_->viewer = new ygCAVEViewer(viewerName.c_str(),true);
    p_->viewer->reset();
    ygSoundServer::initServer();
    frameTime_ = ygGetTime();
    p_->retainRoot = true;
    p_->quitFlag = false;
    p_->editType = "ygWorld";
    p_->saveFrom = "localroot";
    p_->editFrom = "localroot";
    p_->loadFrom = "localroot";
    p_->deleteFrom = "localroot";
    addNetMessage("quit",&p_->quitFlag,YG_NET_BOOL_NEGATE);
    dontSendNetMessage("quit");
    addNetMessage("save",&p_->saveFile,YG_NET_FILE);
    addNetMessage("save",&p_->saveFrom,YG_NET_NODE);
    dontSendNetMessage("save");
    addNetMessage("editType",&p_->editType,YG_NET_CLASS);
    addNetMessage("editType",&p_->editFrom,YG_NET_NODE);
    dontSendNetMessage("editType");
    addNetMessage("load",&p_->loadFile,YG_NET_FILE);
    addNetMessage("load",&p_->loadFrom,YG_NET_NODE);
    dontSendNetMessage("load");
    addNetMessage("delete",&p_->deleteFrom,YG_NET_NODE);
    addNetMessage("delete",&p_->retainRoot,YG_NET_BOOL);
    dontSendNetMessage("delete");
    createDebugNode();
    p_->lastTime = 0.0;
    }


ygWorld::~ygWorld(void)
    {
    p_->root_->removeChild(this);
    ygNet::process();
    while (!p_->messageQueue.empty())
        {
        ygDelayMsg *dm = p_->messageQueue.front();
        p_->messageQueue.pop_front();
        delete dm;
        }
    while (!p_->messageFrameQueue.empty())
        {
        ygDelayMsg *dm = p_->messageFrameQueue.front();
        p_->messageFrameQueue.pop_front();
        delete dm;
        }
    delete p_;
    ygSoundServer::closeServer();
    }


void ygWorld::createRoot(void)
    {
    p_->root_ = ygNodeDB::create("group","root");
    p_->root_->addChild(this);
    p_->pfscene->addChild(p_->root_->pfnode());
    }


void ygWorld::setRoot(ygNode *r)
    {
    p_->root_ = r;
    if (p_->root_)
        {
        p_->root_->addChild(this);
        p_->pfscene->addChild(p_->root_->pfnode());
        }
    }


ygNode* ygWorld::root(void) const
    {
    return p_->root_;
    }


pfScene* ygWorld::pfscene(void) const
    {
    return p_->pfscene;
    }


ygNode* ygWorld::viewer(void) const
    {
    return p_->viewer;
    }


void ygWorld::reset(void)
    {
    frameTime_ = ygGetTime();
    frameNumber_ = 0;
    while (!p_->messageQueue.empty())
        {
        ygDelayMsg *dm = p_->messageQueue.front();
        p_->messageQueue.pop_front();
        delete dm;
        }
    while (!p_->messageFrameQueue.empty())
        {
        ygDelayMsg *dm = p_->messageFrameQueue.front();
        p_->messageFrameQueue.pop_front();
        delete dm;
        }
    //reset viewer
    if (p_->viewer)
        p_->viewer->reset();
    //call resetTree recursively on all localroot children
    ygNode* localroot = ygNodeDB::find("localroot");
    for (int i=0; i < localroot->numChildren(); i++)
        localroot->child(i)->resetTree();
    localroot->ygNode::reset();
    }


void ygWorld::app(void)
    {
    if (debugNode())
        {
        pfVec3 pos;
        if (p_->viewer)
            ((ygCAVEViewer*)(p_->viewer))->getPosition(pos);
        pfMatrix mat;
        mat.makeScale(CAVEFar,CAVEFar,CAVEFar);
        mat.postTrans(mat,pos[0],pos[1],pos[2]);
        pfDCS* dcs = (pfDCS*)debugNode();
        dcs->setMat(mat);
        }
    ygNode::app();
    }


void ygWorld::frame(void)
    {
    float t;
    t = ygGetTime();
    frameDeltaTime_ = t - p_->lastTime;
    if (frameDeltaTime_ <= 0)
        {
        cerr << "WARNING (ygWorld::app): Bogus delta time ("
            << frameDeltaTime_ << "); kludging to .0167\n";
        frameDeltaTime_ = 0.0167;
        }
    //prevent loss of time from long delays
    else if (frameDeltaTime_ > 0.1)
        frameDeltaTime_ = 0.1;
    p_->lastTime = t;
    frameTime_ += frameDeltaTime_;
    frameNumber_++;
    while ((!p_->messageQueue.empty()) &&
        (p_->messageQueue.front()->time <= frameTime_))
        {
        ygDelayMsg *dm = p_->messageQueue.front();
        sendMessage(dm->msg);
        p_->messageQueue.pop_front();
        delete dm;
        }
    while ((!p_->messageFrameQueue.empty()) &&
        (p_->messageFrameQueue.front()->frame <= frameNumber_))
        {
        ygDelayMsg *dm = p_->messageFrameQueue.front();
        sendMessage(dm->msg);
        p_->messageFrameQueue.pop_front();
        delete dm;
        }
    if (p_->root_)
        p_->root_->appTraverse();
    if (p_->viewer)
        p_->viewer->app();
    ygSoundServer::updateServer();
    ygNet::process();
    }


void ygWorld::scheduleMessage(ygMessage& msg,bool queue)
    {
    if (msg.delay() == 0 && !queue)
        sendMessage(msg);
    else
        {
        ygDelayMsg* dm = new ygDelayMsg;
        if (msg.delay() < 0)
            {
            dm->frame = frameNumber_ - (int)msg.delay();
            list<ygDelayMsg*>::iterator i = p_->messageFrameQueue.begin();
            for (; i != p_->messageFrameQueue.end(); ++i)
                if (dm->frame < (*i)->frame)
                   break;
            p_->messageFrameQueue.insert(i,dm);
            }
        else
            {
            dm->time = frameTime_ + msg.delay();
            list<ygDelayMsg*>::iterator i = p_->messageQueue.begin();
            for (; i != p_->messageQueue.end(); ++i)
                if (dm->time < (*i)->time)
                   break;
            p_->messageQueue.insert(i,dm);
            dm->msg = msg;
            }
        dm->msg = msg;
        }
    }


void ygWorld::sendMessage(ygMessage& msg)
    {
    if (p_->sentMessageDebug)
        cout << "ygWorld: sending message " << msg << endl;
    if (!msg.recipient())
        msg.setRecipient(ygNodeDB::find(msg.recipientName(),false));
    if (msg.recipient())
        {
        if (msg.recipient()->isNetMaster())
            msg.recipient()->parseMessage(msg);
        else
            msg.recipient()->sendNetMessage(msg);
        }
    else
        cerr << "ERROR: could not find recipient node for message '" << msg << "'\n";
    }


void ygWorld::message(const ygMessage& msg)
    {
    //quit the application
    if (msg == "quit")
        {
        return;
        }
    //load the given filename below optional node
    else if (msg == "load")
        {
        bool loaded=false;
        if (p_->loadFile.length() > 0)
            {
            p_->saveFile = p_->loadFile;
            if (p_->loadFrom.length() > 0)
                {
                ygNode* node = ygNodeDB::find(p_->loadFrom,false);
                if (node)
                    {
                    loadFromFile(p_->loadFile,node);
                    loaded = true;
                    }
               else
                    msg.error("World","(node not found)");
                }
            else
                {
                loadFromFile(p_->loadFile,this);
                loaded = true;
                }
            }
        else
            msg.error("World","(wrong number of arguments)");
        if (loaded)
            {
            //generate load event for alerts
            eventOccurred("load");
            }
        }
    //save the scene to given filename below optional node
    else if (msg == "save")
        {
        bool saved=false;
        if (p_->saveFile.length() > 0)
            {
            if (p_->saveFrom.length() > 0)
                {
                ygNode* node = ygNodeDB::find(p_->saveFrom,false);
                if (node)
                    {
                    saveToFile(p_->saveFile,node);
                    saved = true;
                    }
                else
                    msg.error("World","(node not found)");
                }
            else
                {
                saveToFile(p_->saveFile,this);
                saved = true;
                }
            }
        else
            msg.error("World","(wrong number of arguments)");
        if (saved)
            {
            //generate save event for alerts
            eventOccurred("save");
            }
        }
    //create a new child node with optional name
    else if (msg == "create")
        {
        ygNode* node=NULL;
        if (msg.numArgs() > 0)
            {
            ygString name;
            ygString type;
            type = msg.stringArg(0);
            if (msg.numArgs() > 1)
                node = ygNodeDB::create(type,msg.stringArg(1));
            else
                {
                ygNodeDB::createValidName(type,name);
                node = ygNodeDB::create(type,name);
                }
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        if (node)
            {
            ygNode* localroot = ygNodeDB::find("localroot");
            localroot->addChild(node);
            node->reset();
            cout << name() << ": created node " << node->name() << endl;;
            }
        }
    //delete the scene below optional node and optional retain root node
    else if (msg == "delete")
        {
        //do not allow localroot to be deleted
        if (p_->deleteFrom == "localroot")
            p_->retainRoot = true;
        if (p_->deleteFrom.length() > 0)
            {
            ygNode* node = ygNodeDB::find(p_->deleteFrom,false);
            if (node)
                deleteFromNode(node,p_->retainRoot);
            else
                msg.error("World","(node not found)");
            }
        else
            deleteFromNode(this,true);
        p_->deleteFrom = "localroot";
        p_->retainRoot = true;
        }
    //set the debug state of a node type from optional node down
    else if (msg == "editType")
        {
        if (p_->editTypeLast.length() > 0)
            {
            ygString msgStr;
            msgStr = "debug(display,false)";
            ygMessage dbgMsg;
            dbgMsg.parseString(msgStr);
            //allways find grid nodes
            if (p_->editFromLast.length() > 0 && p_->editTypeLast != "ygGrid")
                {
                ygNode* node;
                node = ygNodeDB::find(p_->editFromLast,false);
                if (node)
                    sendClassMessage(dbgMsg,p_->editTypeLast,node);
                }
            else
                sendClassMessage(dbgMsg,p_->editTypeLast,this);
            }
        if (p_->editType.length() > 0)
            {
            ygString msgStr;
            msgStr = "debug(display)";
            ygMessage dbgMsg;
            dbgMsg.parseString(msgStr);
            //allways find grid nodes
            if (p_->editFrom.length() > 0 && p_->editType != "ygGrid")
                {
                ygNode* node;
                node = ygNodeDB::find(p_->editFrom,false);
                if (node)
                    sendClassMessage(dbgMsg,p_->editType,node);
                else
                    msg.error("World","(node not found)");
                }
            else
                sendClassMessage(dbgMsg,p_->editType,this);
           }
        p_->editTypeLast = p_->editType;
        p_->editFromLast = p_->editFrom;
        }
    //set the debug state of a node type from optional node down
    else if (msg == "setdebug")
        {
        if (msg.numArgs() > 0)
            {
            ygString *type;
            int i=0;
            ygString message = msg.stringArg(0);
            type = message.nextToken(".",&i);
            ygString *flag;
            flag = message.nextToken(".",&i);
            ygString msgStr = "debug(";
            msgStr += *flag;
            msgStr += ")";
            ygMessage dbgMsg;
            dbgMsg.parseString(msgStr);
            if (msg.numArgs() > 1)
                {
                ygNode* node;
                node = ygNodeDB::find(msg.stringArg(1),false);
                if (node)
                    sendClassMessage(dbgMsg,*type,node);
                else
                    msg.error("World","(node not found)");
                }
            else
                sendClassMessage(dbgMsg,*type,this);
            delete type;
            delete flag;
            }
        else
            msg.error("World","(wrong number of arguments)");
        }
    //unset the debug state of a node type from optional node down
    else if (msg == "unsetdebug")
        {
        if (msg.numArgs() > 0)
            {
            ygString *type;
            int i=0;
            ygString message = msg.stringArg(0);
            type = message.nextToken(".",&i);
            ygString *flag;
            flag = message.nextToken(".",&i);
            ygString msgStr = "debug(";
            msgStr += *flag;
            msgStr += ",false";
            msgStr += ")";
            ygMessage dbgMsg;
            dbgMsg.parseString(msgStr);
            if (msg.numArgs() > 1)
                {
                ygNode* node;
                node = ygNodeDB::find(msg.stringArg(1),false);
                if (node)
                    sendClassMessage(dbgMsg,*type,node);
                else
                    msg.error("World","(node not found)");
                }
            else
                sendClassMessage(dbgMsg,*type,this);
            delete type;
            delete flag;
            }
        else
            msg.error("World","(wrong number of arguments)");
        }
    //graph the scene from optional node and to optional depth
    else if (msg == "graph")
        {
        ygNode* node;
        int depth;
        depth = 99;
        if (msg.numArgs() > 0)
            {
            node = ygNodeDB::find(msg.stringArg(0),false);
            if (msg.numArgs() > 1)
                depth = msg.intArg(1);
            }
        else
            node = ygNodeDB::find("localroot");
        if (node)
            ygPrintSceneGraph(node,depth);
        else
            msg.error("World","(node not found)");
        }
    //assume the viewpoint of another user
    else if (msg == "viewer")
        {
        if (msg.numArgs() > 0)
            {
            ygNode* node;
            node = ygNodeDB::find(msg.stringArg(0));
            if (node && node->isOfClass("ygUser"))
                {
                if (p_->viewer)
                    ((ygCAVEViewer*)p_->viewer)->setUser((ygUser*)node);
                }
            else
                msg.error("World","(user not found)");
            }
        else
            {
            if (p_->viewer)
                ((ygCAVEViewer*)p_->viewer)->setUser();
            }
        }
    else
        ygNode::message(msg);
    }


void ygWorld::addUser(ygUser* user)
    {
    p_->users.push_back(user);
    }


void ygWorld::removeUser(ygUser* user)
    {
    vector<ygUser*>::iterator i = p_->users.begin();
    while ((i != p_->users.end()) && (*i != user))
        ++i;
    if (i != p_->users.end())
        p_->users.erase(i);
    }


int ygWorld::numUsers(void) const
    {
    return p_->users.size();
    }


ygUser* ygWorld::user(int index) const
    {
    if ((index >= 0) && (index < p_->users.size()))
        return p_->users[index];
    else
        return NULL;
    }


void ygWorld::setClip(float near,float far)
    {
    CAVENear = near;
    CAVEFar = far;
    }


void ygWorld::setSkyColor(const pfVec3& sky)
    {
    p_->pfsky->setColor(PFES_CLEAR, sky[0], sky[1], sky[2], 0.0);
    }


void ygWorld::fogOff(void)
    {
    p_->pfsky->setFog(PFES_GENERAL, NULL);
    }


void ygWorld::setFog(int fogtype,const pfVec3& color,float onset,float opaque)
    {
    p_->pffog->setFogType(fogtype);
    p_->pffog->setColor(color[0],color[1],color[2]);
    p_->pffog->setRange(onset,opaque);
    p_->pfsky->setFog(PFES_GENERAL, p_->pffog);
    }


void ygWorld::deleteFromNode(ygNode* node,bool keep)
    {
    while (node->numChildren() > 0)
        deleteTree(node->child(0));
    if (!keep)
        {
        if (node->parent())
            node->parent()->removeChild(node);
        delete node;
        }
    if (p_->viewer)
        p_->viewer->reset();
    appTraverse();
    }


void ygWorld::deleteLocalScene(void)
    {
    while (numChildren() > 0)
        deleteTree(child(0));
    }


void ygWorld::deleteTree(ygNode* node)
    {
    if (!node)
        return;
    while (node->numChildren() > 0)
        deleteTree(node->child(0));
    if (node->parent())
        node->parent()->removeChild(node);
    delete node;
    }


void ygWorld::saveToFile(const ygString& filename,ygNode* node)
    {
    ofstream fout;
    fout.open(filename.c_str());
    for (int i=0; i < node->numChildren(); i++)
        saveTree(fout,node->child(i),0);
    fout.close();
    }


void ygWorld::loadFromFile(const ygString& filename,ygNode* node)
    {
    ygParser parser;
    parser.parseFile(filename,node);
    if (p_->viewer)
        p_->viewer->reset();
    }


void ygWorld::saveTree(ostream& s,ygNode* node, int depth)
    {
    if (node)
        {
        int i;
        node->printMessages(s,depth,true);
        depth++;
        if (node->numChildren() && node->className() != "ygScene")
            {
            for (i=0; i < depth; i++)
                s << "    ";
            s << "{" << endl;
            for (i=0; i < node->numChildren(); i++)
                saveTree(s,node->child(i),depth);
            for (i=0; i < depth; i++)
                s << "    ";
            s << "}" << endl;
            }
        }
    }


void ygWorld::sendClassMessage(const ygMessage& msg,const ygString& type,ygNode* node)
    {
    if (node)
        {
        if (type == "*" || node->isOfClass(type))
            node->parseMessage(msg);
        for (int i=0; i < node->numChildren(); i++)
            sendClassMessage(msg,type,node->child(i));
        }
    }


bool ygWorld::quit(void)
    {
    return p_->quitFlag;
    }


void ygWorld::updateDebug(void)
    {        
    if (debugDisplay())
        debugNode()->setTravMask(PFTRAV_ISECT, YG_ISECT_ALLMASK | YG_ISECT_DEBUGMASK, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    else
        debugNode()->setTravMask(PFTRAV_ISECT, 0, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    }


void ygWorld::createDebugNode(void)
    {
    pfGeode* geode = new pfGeode;
    if (!geode)
        {
        cerr << "ERROR: ygNode::createDebugNode: failed to create pfGeode\n";
        return;
        }
    pfGeoSet *gset = makeSphereGeoset();
    geode->addGSet(gset);  
    geode->setName("world");
    pfDCS* dcs = new pfDCS;
    dcs->addChild(geode);
    dcs->setTravMask(PFTRAV_ISECT, YG_ISECT_ALLMASK | YG_ISECT_DEBUGMASK, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    dcs->setTravMask(PFTRAV_DRAW, 0, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    setDebugNode((pfGroup*)dcs);
    }

#include <Performer/pfdu.h>

pfGeoSet* ygWorld::makeSphereGeoset(void)
    {
    pfGeoSet *gset;
    pfMatrix matrix;
    gset = pfdNewSphere(32,pfGetSharedArena());
    pfVec4* color = (pfVec4 *) pfMalloc(sizeof(pfVec4),pfGetSharedArena());
    color->set(1,1,1,1);
    gset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);
    gset->setGState(new pfGeoState);
    gset->setLineWidth(1.0);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    gset->getGState()->setMode(PFSTATE_ENWIREFRAME,1);
    matrix.makeScale(0.5,0.5,0.5);
    pfdXformGSet(gset,matrix);
    return gset;
    }
